/////////////////////////////////////////////////////////////////////////////////////////////
//		 Bluetooth-I/O für Arduino Intervall-Timer
//
//       17.08.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////
// UTF-8!

#include "IntervallTimer.h"

#if BLUESERIAL_IN_SOFTWARE
    BufferedSoftwareSerial blueSerial(blueRX, blueTX);
#endif
    bool    debug;

//
// Set input mode:
//  0: Accept command characters, optionally preceeded by an unsigned integer number.
//  1: Read floating point PID coefficient, terminated by 'P', 'I', or 'D'.
//     (float number format: "+/-1.2e-3")
//
void IntervallTimer::EnterInputMode(int mode)
{
  inputMode = mode,
  numDigits = 0;
}

//
// Neue Bluetooth-Eingaben auswerten
// 
// Rückgabe: true wenn da, sonst false.
//
bool IntervallTimer::CheckBlue()
{
  if (!blueSerial.available()) {	// Zeichen über BT da?
    return false;			//  nein
  }
  int8_t z = blueSerial.read();		//  ja: abholen
  if (z == 10  ||  z == 13) {		//   cr, lf komplett ignorieren,
    sentNewline++;			//    aber protokollieren, ob jedes Zeichen eines 
    return false;			//    mitbekommt (wenn entsprechend eingestellt)
  }
  sentNewline = 0;
  z = toupper(z);
  if (inputMode != 0) {			// Zahleneingabe für PID?
    // numDigits = Anzahl Zeichen
    // erstes Z. darf + oder - sein
    // alles außer Digit beendet den Mode

    if ((isDigit(z) || z=='-' || z=='+' || z=='.' || z=='E')) { // Ziffer etc.?
      numberBuffer[numDigits] = z;
      numDigits++;
      if (numDigits < ELEMENTS(numberBuffer)) {	      //  und noch mehr Platz
	return true;
      }
    }
    else {
      numberBuffer[numDigits] = 0;
      float f;
      char* p;
      f = strtod(numberBuffer, &p);	  // (spart ggü. sscanf() knapp 1 KB)
      if (numDigits > 0 && *p == 0) {	  // gültige Zahl?
#ifdef COOLER_PWM
	if (inputMode == 1) {		  // PID-Koeff?
	  switch (z) {
	    case 'P':
	      KpKiKd[0] = f;
	    break;
	    case 'I':
	      KpKiKd[1] = f;
	    break;
	    case 'D':
	      KpKiKd[2] = f;
	    break;
	    default:
	      SendErrorBlue();
	      EnterInputMode(0);
	      return true;
	  }
	  EnterNewLine(); PrintPID(KpKiKd); PrintNewLine();
	  EnterInputMode(0);
  	  return true;
	}
	else if (inputMode == 3 && z == 'K') {		  // Ziel-Temp.?
	  EnterNewLine(); 
	  StartPIDT(f*10);	   
	  EnterInputMode(0);
  	  return true;
	}
	else 
#endif
	if (inputMode == 2  && z == 'T') {		  // Temp.-Korr?
	  tempCorr = int(f * 10);
	  EnterNewLineLn(PSTR("Temperturkorrektur übernommen, kann mit 'E' gespeichert werden."));
	  EnterInputMode(0);
  	  return true;
	}
      }
    }
    SendErrorBlue();
    EnterInputMode(0);
    return true;
  }
  if (z == 'R') {    // Reset
    Reset();
    EnterNewLine();	  // frische Zeile
    return true;
  }
  if (z >= '0'  &&  z <= '9') {		// Zifferneingabe?
    if (HasNumber()) {
      numberInput = numberInput * 10 + (z - '0');
    }
    else {
      numberInput = (z - '0');
    }
    numDigits++;
    return true;
  }
#if defined(COOLER_PWM) || defined(COOLER_POWER)
# define NUMCMNDS "SMLNP%ECK"
#else
# define NUMCMNDS "SMLNP%E"
#endif
  if (strchr(NUMCMNDS, z) != 0) {	// Befehle mit Zahlenangabe:
    //
    // Zuerst prüfen, ob ein Zeichen, das eine Zifferneingabe übernimmt,
    //  vorliegt und ggf. auswerten:
    //
    if (z == 'S') {    // Sekunden
      if (!HasNumber()) {
	ShowHelpOn_s();
      }
      else if (numberInput == 0) {
	speedIndex = 0;	    // schnelle Auslösung
	SendInfoBlue(1);
      }
      else if (-numberInput < 0) {  // 1..-minInt
	speedIndex = -numberInput;
	SendInfoBlue(1);
      }
      else {
	SendErrorBlue();
      }
    }
    else if (z == 'M') {    // Minuten
      if (!HasNumber()) {
	ShowHelpOn_m();
      }
      else if (numberInput == 0) {
	speedIndex = numberInput;	    // schnelle Auslösung
	SendInfoBlue(1);
      }
      else if (numberInput <= (0x7FFF / 60)) {
	speedIndex = -(numberInput * 60);
	SendInfoBlue(1);
      }
      else {
	SendErrorBlue();
      }
    }
    else if (z == 'L') {    // Limit in Minuten
      if (!HasNumber()) {
	ShowHelpOn_l();
      }
      else SetLimit(numberInput);
    }
    else if (z == 'N') {    // Limit = Anzahl Frames
      if (!HasNumber()) {
	ShowHelpOn_n();
      }
      else if (numberInput == 0) {
	SetLimit(0);	  // endlos
      }
      else {
	SetLimit((GetSpeed() * numberInput + 59999L) / 60000L);	// Gesamtzeit (aufgerundet) in msec --> min
      }
    }
    else if (z == 'P') {    // Pause in Zehnteln
      if (!HasNumber()) {
	ShowHelpOn_p();
      }
      else if (numberInput > 0) {
	bulbBreak_ms = numberInput * 100L;  // --> millis
	//
	// min. Bulb-Pause einhalten: bei zu kleinen Pausen startet
	//  nur jede zweite Auslösung!
	/* Bei Bedarf durch Reset einstellen:
	if (bulbBreak_ms < bulbBreak_min) {   // Minimalwert
	  bulbBreak_ms = bulbBreak_min;
	}
	*/
	SendInfoBlue(1);
      }
      else {
	SendErrorBlue();
      }
    }
    else if (z == '%') {    // Belichtungszeit variieren, in %
      if (!HasNumber()) {
	ShowHelpOn_pcnt();
      }
      else if (numberInput < 100) {
	speedDither = numberInput;
	speedOffset = 0;
	SendInfoBlue(1);
      }
      else {
	SendErrorBlue();
      }
    }
    else if (z == 'E') {    // Einstellungen
      if (!HasNumber()) {
	EnterNewLine(); StartHelp(settingsHelpPage); 
      }
      else switch (numberInput) {
	case 0:			// aktuelle Settings anzeigen
	  ShowSettings();
	break;
	case 1:
	  ResetSettings(false);
	  EnterNewLine(PSTR("Einstellungen auf Basiswerte zurückgesetzt."));
          SendInfoBlue(1);
	break; 
	case 2:
	  if (ResetSettings(true)) {
	    EnterNewLine(PSTR("Einstellungen aus EEPROM geladen."));
  	    SendInfoBlue(1);
	  }
	  else {
	    EnterNewLine(PSTR("Es sind keine Einstellungen im EEPROM gespeichert."));
	  }
	break;
	case 3:			// Settings in EEPROM speichern
	  SaveSettingstoEEPROM();
	  EnterNewLine(PSTR("Einstellungen in EEPROM gespeichert."));
	break; 
	case 4:			// Settings aus EEPROM löschen
	  ClearSettingsfromEEPROM();
	  EnterNewLine(PSTR("Einstellungen aus EEPROM gelöscht."));
          SendInfoBlue(1);
	break; 
	default:	SendErrorBlue();
      }
    }
#if defined(COOLER_PWM) || defined(COOLER_POWER)
    else if (z == 'K') {    // Kühlungsoptionen: 0=aus 1=min, 2=low, 3=std, 4=hi, 5=max, 1xxx=xxx, 8=geregelt, 9=geregelt nach 10 min
      if (!HasNumber()) {
	EnterNewLine(); 
	#ifdef COOLER_PWM
	StartHelp(calibHelpPage); 
	#else
	ShowHelpOn_kh();
	#endif
      }
      else switch (numberInput) {
	case 0:	
	case 1:	
	case 2:	
	case 3:	
	case 4:	
	case 5:	StopPID(); SetCoolingMode(numberInput); SendInfoBlue(1); break;
	case 7:	bluePrint_P(PSTR(" Ziel-Temperatur eingeben, mit 'K' abschließen: "));
		EnterInputMode(3);
		break;
	case 8:	StartPID(0);   EnterNewLine(); break;
	case 9:	StartPID(600); EnterNewLine(PSTR("Temperaturregelung startet in 10 Minuten.")); break;
	#if defined(HEATER_POWER)
	case 21: 
	# if defined(HEATER_PWM) || defined(HEATER_SOFT_PWM)
	case 22: 
	case 23: 
	case 24: 
	case 25: 
	# endif
	case 20: SetHeatingMode(numberInput-20); SendInfoBlue(1);break;
	#endif
	default: 
	  if (numberInput >= 1000  &&  numberInput <= 1255) {
	    StopPID(); 
	    SetCoolingLevel(numberInput - 1000);
	    SendInfoBlue(1);
	  }
	  else {
  	    SendErrorBlue(); 
	  }
	  break;
      }
    }
#endif
#ifdef COOLER_PWM
    else if (z == 'C') {    // Kalibrierung
      if (!HasNumber()) {
	if (IsCalibPID()) {    // "PID-Kalibirerung aus" geht auch ohne
	  UpdatePIDCalib(true);	  // stop
	}
	else {
	  EnterNewLine(); StartHelp(calibHelpPage); 
	}
      }
      else switch (numberInput) {
	case 0:	EnterNewLine(); PrintPID(KpKiKd); PrintNewLine(); break; // Print PID
	case 1:			        // PID auf Basiseinstellung
	  ResetPID(false);
	  EnterNewLine(PSTR("PID-Koeffizienten auf Basiswerte zurückgesetzt."));
	break;
	case 2:				// Reset PID
	  if (ResetPID(true)) {
	    EnterNewLine(PSTR("PID-Koeffizienten aus EEPROM geladen."));
	  }
	  else {
	    EnterNewLine(PSTR("Es sind keine PID-Koeffizienten im EEPROM gespeichert."));
	  }
	break;
	case 3:			// Settings in EEPROM speichern
	  SavePIDtoEEPROM();
	  EnterNewLine(PSTR("PID-Koeffizienten in EEPROM gespeichert."));
	break; 
	case 4:				// PID auf Basiseinstellung, aus EEPROM löschen
	  ClearPIDfromEEPROM();
	  EnterNewLine(PSTR("PID-Koeffizienten im EEPROM gelöscht."));
	break; 
	case 5:		    	// Enter PID
	  bluePrint_P(PSTR(" PID-Koeffizient eingeben, mit P/I/D abschließen: "));
	  EnterInputMode(1);
	break; 
	default:		        // Start PID Calib
	  EnterNewLine(); 
	  CalibPID(numberInput);
	break; 
      }
    }
#endif
  }
  else if (HasNumber()) {			// die weiteren Kommandos sind nur erlaubt, wenn keine Zifferneingabe vorausgeht
    EnterNewLine(PSTR("Zahleneingabe hier nicht erlaubt!"));
  }
#if defined(TEMPSENS)
  else if (z == 'T') {
    bluePrint_P(PSTR(" Temperatur-Korrektur eingeben, mit 'T' abschließen: "));
    EnterInputMode(2);
  }
#endif
  else if (z == '!'  ||  z == '#') {
    if (shootState == stIdle) {	        // aus?
      if (z == '!') {			// neue Auslösung:
	speedOffset = 0;		// Dither-Start
	StartFrame();			//  ab jetzt Shoot wie gewählt wiederholen
	SetShootLimit();		// Limit-Zeit eintragen
      }
      else {				// '#'
  	shootState = stShootBulb;	
	SetShootLine(true);		//   los!
	endShoot = millis();		//   ab jetzt
	nextUpdate = endShoot + 1000;
      }
    }
    else {				// schon an:
      SetShootLine(false);		//  dann abschalten
      if (shootState == stShootBulb) {	// Kamera auf Bulb?
	if (quiet) {			//  dann nur wenn quiet:
          bluePrint_P(PSTR(" ("));
	  blueSerial.print((int)((millis() - endShoot)/1000));
          bluePrintln_P(PSTR("s)"));
	}
      }
      else {
	if (!IsFastSpeed()) {
	  int32_t dauer = GetSpeed() - (endShoot - millis());
	  if (dauer > 0) {		// negativ, wenn t verkürzt wurde
	    bluePrint_P(PSTR(" ("));
	    blueSerial.print((int)(dauer / 1000));
  	    bluePrintln_P(PSTR("s)"));
	  }
	}
      }
      shootState = stIdle;
      EnterNewLine();
    }
  }
  else if (z == '+'  ||  z == '-') {	// Belichtungsschritt:
    if (speedIndex < 0) {		// bisher direkt in Sekunden angegeben?
      uint32_t spd_ms = GetSpeed();	//  in msec in der Tabelle suchen
      if (z == '+') {			//   von unten her, wenn '+':
	speedIndex = 1;
	do {
	  if (GetSpeed() > spd_ms) {
	    break;			//   fertig, wenn neue Einstellung den alten Wert überschritten hat
	  }
	  speedIndex++;
	} while (speedIndex < GetNumSpeeds());
      }
      else {	      			//   von oben her, wenn '-':
	speedIndex = GetNumSpeeds() - 1;
	do {
	  if (GetSpeed() < spd_ms) {
	    break;			//   fertig, wenn neue Einstellung den alten Wert unterschritten hat
	  }
	  speedIndex--;
	} while (speedIndex > 1);
      }
    }
    else {				// Belichtung nach Tabellenwert: um 1 Schritt ändern
      if (z == '+') {
	if (speedIndex < GetNumSpeeds() - 1) {	  //  noch nicht durch?
	  speedIndex++;
	}
      }
      else {
	if (speedIndex > 0) {
	  speedIndex--;
	}
      }
    }
    SendInfoBlue(1);
  }
  else if (z == '>'  ||  z == '<') {	// Zeitbegrenzung
    if (limitIndex < 0) {		// bisher direkt in Minuten angegeben?
      uint32_t lim_ms = GetLimit();	//  bisheriges Limit in msec
      // 
      // Passenden Wert in der Tabelle suchen:
      // Zur Verkleinerung ('<') von oben her (d.h. in der Tabelle von 
      //  links nach rechts) suchen, bis der Tabellenwert kleiner als der
      //  bisherige Wert ist.
      // Zur Vergrößerung ('>') von unten her (d.h. in der Tabelle von hinten).
      // 
      if (z == '<') {			//  '<': von oben her:
	limitIndex = 1;			//   ([0] = 0 == endlos!)
	do {
	  if (GetLimit() < lim_ms) {
	    break;			//   fertig, wenn neue Einstellung den alten Wert unterschritten hat
	  }
	  limitIndex++;
	} while (limitIndex < GetNumLimits());
      }
      else {	      			//   '>': von unten her:
	limitIndex = GetNumLimits() - 1;//   letzter Wert
	do {
	  if (GetLimit() > lim_ms) {
	    break;			//   fertig, wenn neue Einstellung den alten Wert überschritten hat
	  }
	  limitIndex--;
	} while (limitIndex > 1);	// ([0] = 0 == endlos!)
      }
    }
    else {				// Belichtung nach Tabellenwert: um 1 Schritt ändern
      if (z == '>') {		// erhöhen:
	if (limitIndex > 0) {	//  noch nicht durch?
	  limitIndex--;
	}
      }
      else {
	if (limitIndex < GetNumLimits() - 1) {  //  noch nicht durch?
	  limitIndex++;
	}
      }
    }
    SetShootLimit();		  	  // aktuelle Limit-Zeit updaten (wird bei Auslösung erneuert)
    SendInfoBlue(2);
  }
  else if (z == '?') {    // Info
    SendInfoBlue(-1);
  }
  else if (z == 'U') {    // Updates
    quiet = !quiet;
    if (shootState == stIdle) {
      SendInfoBlue(-1);
    }
    else {
      EnterNewLine();		  // frische Zeile
    }
  }
  else if (z == 'D') {    // Debug
    debug = !debug;
    if (shootState == stIdle) {
      SendInfoBlue(-1);
    }
    else {
      EnterNewLine();		  // frische Zeile
    }
  }
  else if (z == 'H') {    // Hilfe
    StartHelp();
  }
  else {
    SendErrorBlue();
  }
  numDigits = 0;
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
//
// String direkt aus dem PROGMEM an blueSerial ausgeben
//
void bluePrint_P(PGM_P s)
{
  char c;
  while ((c = pgm_read_byte(s)) != 0) {
    blueSerial.write(c); s++;
  }
}

// .. mit newline
inline
void bluePrintln_P(PGM_P s)	   
{
  bluePrint_P(s); 
  blueSerial.println();
}


//
// Eine mit 10^dcms skalierte Zahl als Dezimalzahl an blueSerial ausgeben.
// Die Zahl der Dezimalen wird um die Zahl der Vorkommastellen - 1
//  reduziert auf minimal eine.
//
void bluePrintDecimals(int val, int dcms)
{
  if (val < 0) {
    blueSerial.write('-');
    val = -val;
  }
  char bf[30];
  int n = 0;
  while (val != 0) {    // Ziffernfolge von rechts nach links:
    int r = val;
    val = val / 10;
    bf[n++] = (char)(r - val*10 + '0');
  }
  while (n <= dcms) {	// mit 0 auffüllen bis zur ersten Vorkommastelle
    bf[n++] = '0';
  }
  int i = n;
  while (i > 0) {	// Ziffern rückwärts ausgeben:
    if (i == dcms) {	// noch genau dcms Ziffern?
      blueSerial.write('.');
      // n-i - 1 = Zahl der angez. Vorkommastellen - 1
      // Von den restlichen i Stellen, den Dezimalen, abziehen
      //  --> i - (n-i-1) = 2i - n + 1
      //  aber mindestens eine:
      //
      i = max(1, 2*i - n + 1);
    }
    i--;
    blueSerial.write(bf[i]);
  }
}

////////////////////////////////////////////////////////////////////////////////////
#if BLUESERIAL_IN_SOFTWARE
//
// Zahl in den Puffer schreiben
//
void BufferedSoftwareSerial::print(long l)
{
  if (l < 0) {
    write('-');
    l = -l;
  }
  print((unsigned long)l);
}

void BufferedSoftwareSerial::print(unsigned long l)
{
  if (l == 0) {
    write('0');
    return;
  }
  char bf[30];
  int x = 0;
  while (l != 0) {    // Ziffernfolge von rechts nach links:
    long r = l;
    l = l / 10;
    bf[x++] = (char)(r - l*10 + '0');
  }
  while (x > 0) {     // Ziffern rückwärts ausgeben:
    x--;
    write(bf[x]);
  }
}

void BufferedSoftwareSerial::print(int l)
{
  if (l < 0) {
    write('-');
    l = -l;
  }
  print((unsigned int)l);
}

void BufferedSoftwareSerial::print(unsigned int l)
{
  if (l == 0) {
    write('0');
    return;
  }
  char bf[30];
  int x = 0;
  while (l != 0) {    // Ziffernfolge von rechts nach links:
    int r = l;
    l = l / 10;
    bf[x++] = (char)(r - l*10 + '0');
  }
  while (x > 0) {     // Ziffern rückwärts ausgeben:
    x--;
    write(bf[x]);
  }
}

#endif //BLUESERIAL_IN_SOFTWARE

