/////////////////////////////////////////////////////////////////////////////////////////////
//		 Arduino Intervall-Timer mit Bluetooth-Modul
//
//       17.05.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////
#ifndef IntervallTimer_h_
#define IntervallTimer_h_

#define ELEMENTS(t)	(sizeof(t)/sizeof((t)[0]))

#include <arduino.h>
#include "pins.h"
#include "digitalWriteFast.h"

#if BLUESERIAL_IN_SOFTWARE
// Anbindung des Bluetooth-Moduls (HC-05) über Software-Serial() mit Sende-Puffer:
//
#include "BufferedSoftwareSerial.h"
extern BufferedSoftwareSerial blueSerial;
#else
#define blueSerial  Serial
#endif

#define blueBaud    9600	// Baudrate zur Anbindung des Bluetooth-Moduls


extern bool debug;

inline void bluePrint(int i)		  { blueSerial.print(i); }
inline void bluePrintln(int i)		  { blueSerial.println(i); }
       void bluePrint_P(PGM_P s);
       void bluePrintln_P(PGM_P s);
       void bluePrintDecimals(int val, int n);

inline void dbgPrint(int i)		  { if (debug) bluePrint(i); }
inline void dbgPrintln(int i)		  { if (debug) bluePrintln(i); }
inline void dbgPrintDecimals(int i, int n){ if (debug) bluePrintDecimals(i, n); }
inline void dbgPrint_P(PGM_P s)		  { if (debug) bluePrint_P(s); }
inline void dbgPrintln_P(PGM_P s)	  { if (debug) bluePrintln_P(s); }


#include "cooling.h"
#include "StepResponse.h"

///////////////////////////////////////////////////////////////

// Faktoren zur Umrechnung von Zeiten in die Basiseinheit Millisekunde
// Vorsicht vor Konflikt mit <ms> in den Shutteremus, das ms in µs umrechnet!
//
#define sec *1000L     		// Zeit von Sekunden in die Basiseinheit (ms) umrechnen
#define Min *60000L    		// Zeit von Minuten in die Basiseinheit (ms) umrechnen
#define h   *3600000L  		// Zeit von Stunden in die Basiseinheit (ms) umrechnen

struct ivtSettings {
    void    SaveSettingstoEEPROM();
    bool    LoadSettingsfromEEPROM();
    void    ClearSettingsfromEEPROM();
  
  protected:
    int	    speedIndex;	      // Timer-Einstellung
    int	    speedDither;      // Timer-Variation
    int	    limitIndex;	      // Wiederholdauer-Einstellung
    int32_t bulbBreak_ms;     // Signal-Unterbrechung zwischen Bulb-Auslösungen
    bool    quiet;
    int	    coolingLevel;
    int	    heatingLevel;
    int	    tempCorr;
};

class IntervallTimer : public ivtSettings {
  public:
	    IntervallTimer();
	   ~IntervallTimer();

    void    Setup();
    bool    CheckBlue();
    void    Advance();
    bool    SendUpdate();
    int	    UpdatePID();
    bool    SendNextBlue();

  protected:
    #ifdef COOLER_PWM
    void    StartPID(int afterSeconds);
    void    StartPIDT(float ziel);
    void    InitPID(int tempX10);
    void    StopPID();
    bool    CalibPID(int step);
    void    DeletePID();
    bool    IsPIDActive() const;
    bool    IsCalibPID() const;
    int	    UpdatePIDCalib(bool stop);
    bool    ResetPID(bool fromEEPROM);
   static void PrintPID(float*);
    void    SavePIDtoEEPROM();
    bool    LoadPIDfromEEPROM();
    void    ClearPIDfromEEPROM();
    void    ContinueCalib();
    void    SetCoolingLevel(int level)	    {coolingLevel = level; analogWrite(COOLER_PWM, coolingLevel);}
    void    SetHeatingMode(int mode);
    void    SetCoolingMode(int mode);
    #endif

    void    Reset();
    bool    ResetSettings(bool fromEEPROM);
    void    SetShootLimit();
    void    SetShootLine(bool active);
    void    RandomSpeedOffset();
    void    StartFrame();
    uint32_t GetSpeed() const;
    uint32_t GetLimit() const;
    bool    SetLimit(int32_t minutes);
    int	    GetNumSpeeds() const;
    int	    GetNumLimits() const;
    bool    IsFastSpeed() const	    {return speedIndex == 0;}	      // Timer für schnelle Auslösung wenn nicht Bulb

    void    EnterInputMode(int mode);
    bool    HasNumber() const	    {return numDigits != 0;}

    void    SendInfoBlue(int which);
    void    PrintNewLine();
    void    EnterNewLine();
    void    EnterNewLine(PGM_P s);
    void    EnterNewLineLn(PGM_P s);
    void    SendErrorBlue();
    
    void    PrintMore(bool more);
    void    PrintCoolerStatus();
    void    PrintHeaterStatus();
    void    PrintSpeed();
    void    ShowSettings();
    void    ShowHelpOn_s();
    void    ShowHelpOn_m();
    void    ShowHelpOn_l();
    void    ShowHelpOn_n();
    void    ShowHelpOn_p();
    void    ShowHelpOn_pcnt();
    void    ShowHelpOn_stat();
    #if !defined(COOLER_PWM) && defined(COOLER_POWER)
    void    ShowHelpOn_kh();
    #endif
    void    ShowHelpOn_h();
    int     SendHelp(int what);
    void    StartHelp(int page = 1)  	    {helpPage = page;}
    enum {calibHelpPage = 100, settingsHelpPage = 110};

  private:
    uint32_t endShoot;		// Beginn der laufenden Auslösung
    uint32_t shootLimit;	// Ende der laufenden Auslösungsserie
    int	    shootState;	      // Status der Auslösung
    int	    numFrames;
    const int32_t bulbBreak_min = 200;    // min. Bulb-Pause: bei zu kleinen Pausen startet nur jede zweite Auslösung
    const int32_t shootBreak_ms = 100;    // Signal-Unterbrechung zwischen Auslösungen, 2-fach für min. Bulb-Pausen
    int32_t nextUpdate;
    int	    speedOffset;      // aktuelle Timer-Variation
    int	    helpPage;
    int	    calibState;
    unsigned numDigits;
    char    numberBuffer[10];
    int	    numberInput;
    int	    inputMode;
    int	    sentNewline;

#ifdef TEMPSENS
    int	    ReadTemperatureX10();
#endif
#ifdef COOLER_PWM
    float   KpKiKd[3];
    int	    PID_delay;
    int	    PID_SetTemp;
    int     PID_error;
    bool    PID_coolingOff;
    float   PID_i;
    #define RSCAL 1	      // additional scaling of ReadTemperature() for PID calibration, 1 or 4
    #if (RSCAL==4)
    #define ReadTemperature_xS	ReadTemperatureX40
    const int initialResponseDelta_xS = 2; // 0.05 °C Schritte für die Kalibrierung mit T*40
    #else // RSCAL = 1
    #define ReadTemperature_xS	ReadTemperatureX10
    const int initialResponseDelta_xS = 1; // 0.1 °C Schritte für die Kalibrierung mit T*10
    #endif
    StepResponse<100, RSCAL>* PIDCal;
   uint32_t PID_prevTime;
    int	    PID_calCnt;
# define PID_EEPROM_ADDR	0	// address of PID coeffs saved in EEPROM
# define SETTINGS_EEPROM_ADDR (PID_EEPROM_ADDR+4+sizeof(IntervallTimer::KpKiKd))	// address of settings saved in EEPROM
#else
# define SETTINGS_EEPROM_ADDR 0	// address of settings saved in EEPROM
#endif

  public:
#ifdef COOLER_PWM
    static int GetPIDEEPROMAddress()	    {return PID_EEPROM_ADDR;}
#endif
    static int GetSettingsEEPROMAddress()   {return SETTINGS_EEPROM_ADDR;}

    enum {
      stIdle,
      stShoot,
      stShootPaused,
      stShootBulb,
    };
};


inline
IntervallTimer::IntervallTimer()
{
#ifdef COOLER_PWM
  PIDCal = 0;
#endif
  SetShootLine(false);		// (mit digitalWriteFast() setzen!)
  pinMode(SHOOT_LINE, OUTPUT);
}

inline
IntervallTimer::~IntervallTimer()
{
#ifdef COOLER_PWM
  DeletePID();
#endif
}

//
// Initialisierung, die beim Erzeugen eines globalen Objekts noch
//  nicht durchgeführt werden kann.
//
inline 
void IntervallTimer::Setup() 
{  
  blueSerial.begin(blueBaud);	  //  Serial wird erst nach globalen Objekten eingerichtet!
#ifdef HELLO_MSG
  bluePrintln_P(PSTR(HELLO_MSG));
#endif
  setupTemperatureControl();
  Reset();			  // Kühlung erst jetzt verfügbar
}

// Nächstes gepuffertes Bluetooth-Zeichen senden. Wenn Puffer leer ist,
//  zurückgestellte Aktionen weiterführen (Hilfe, PID-Calib).
// 
// Muss von der globalen Ablaufsteuerung regelmäßig aufgerufen werden!
// 
// Gibt false zurück, wenn nichts zu tun ist.
// 
inline 
bool IntervallTimer::SendNextBlue() 
{
#if BLUESERIAL_IN_SOFTWARE
  if (blueSerial.SendNextChar()) {    // nächstes Zeichen senden
    return true;		      // gesendet
  }
#endif
  //
  // Puffer war leer: ist was zurückgestellt?
  // 
  if (helpPage != 0) {
    helpPage = SendHelp(helpPage);    // nächste Charge an Zeichen in den Sendepuffer
    return true;
  }
#ifdef COOLER_PWM
  if (calibState != 0) {
    ContinueCalib();		      // nächste Charge an Zeichen in den Sendepuffer
    return true;
  }
#endif
  return false;			      // alles erledigt
}


// Endezeitpunkt der Auslösungsserie berechnen und eintragen
inline 
void IntervallTimer::SetShootLimit() 
{
  shootLimit = GetLimit();
  if (shootLimit != 0) {	    // nicht endlos?
    shootLimit += millis();	    // Endzeit
    if (shootLimit == 0) {
      shootLimit = 1;		    // nicht endlos
    }
  }
}

// Auslösung (de-)aktivieren
inline 
void IntervallTimer::SetShootLine(bool active) 
{
  digitalWriteFast(SHOOT_LINE, active ? 0 : 1);
}


// 
// Neue zufällige Abweichung von der Belichtungszeit festlegen
//
inline 
void IntervallTimer::RandomSpeedOffset()
{
  // GetSpeed() * random(speedDither / 100) mit besserer Auflösung:
  speedOffset = ((GetSpeed() >> 6)
		 * random(speedDither * ((1<<12)/100))	    // vorab schon durch 100 teilen
		) >> 6;
}


// Auslösesignal anlegen und Status auf stShoot setzen.
// In endShoot die Zeit für das Ende der Auslösung eintragen.
//
inline 
void IntervallTimer::StartFrame()
{
  endShoot = millis() + GetSpeed();
  SetShootLine(true);		//   los!
  if (speedIndex != 0  &&  speedDither > 0) {	// Bulb dithern?
    endShoot += speedOffset;		// Offset zugeben
    if (speedOffset > 0) {		// nächstes mal negativ
      speedOffset = -speedOffset;
    }
    else {
      RandomSpeedOffset();
    }
  }
  shootState = stShoot;
  nextUpdate = -1;
  numFrames++;
}


///////////////////////////////////////////////////////////////
inline bool timeReached(uint32_t t) __attribute__((always_inline));
inline bool timeReached(uint32_t t, uint32_t now) __attribute__((always_inline));

// Prüfen, ob der Zeitpunkt t (in msec) erreicht ist
inline bool timeReached(uint32_t t)
{
  return timeReached(t, millis());
}

// Prüfen, ob der Zeitpunkt t (in msec) now erreicht hat.
inline bool timeReached(uint32_t t, uint32_t now)
{
  return (long)(t - now) < 0;	// auch ok für Überlauf von milliTime bis 0x7FFFFFFF
}

// Prüfen, ob zum aktuellen Auslöse-Zustand Aktionen erforderlich sind.
// - Zeit ist rum: Pause
// - Pause zu ende: nächste Auslösung
//
inline 
void IntervallTimer::Advance()
{
  if (shootState == stShootPaused) {	// beendete Auslösung:
    if (timeReached(endShoot)) {	// Pause zu Ende?
      if (shootLimit != 0  &&  timeReached(shootLimit)) {	// Zeitlimit erreicht?
        shootState = stIdle;
	blueSerial.println("...Limit");
      }
      else {
	StartFrame();
        nextUpdate = -2;
      }
    }
  }
  else if (shootState == stShoot) {     // laufende Auslösung: 
    if (timeReached(endShoot)) {	// Ende erreicht?
      // Auslösesignal beenden und Status auf stShootPaused setzen.
      // In endShoot die Zeit für die nächste Auslösung (nach delay) eintragen.
      //
      SetShootLine(false);
      endShoot = millis() + (IsFastSpeed() ? shootBreak_ms : bulbBreak_ms);
      shootState = stShootPaused;
    }
  }
}

//
// Info über Bluetooth senden
// 
// Rückgabe: true, wenn alles erledigt ist
//
inline 
bool IntervallTimer::SendUpdate()
{
  if (quiet) {
    return true;
  }
  uint32_t now = millis();
  if (shootState == stShootBulb) {	// Kamera auf Bulb
    if (timeReached(nextUpdate, now)) {
      blueSerial.write(' ');
      blueSerial.print((int)((now - endShoot)/1000));
      do {
	nextUpdate += 1000;
      } while (timeReached(nextUpdate, now));  // weiter bis aktuelle Zeit erreicht (falls zuvor quiet!)
    }
  }
  else while (shootState == stShoot) { 
    static uint32_t deltaStern;
    if (nextUpdate < 0) {		  // neue Auslösung?
      if (IsFastSpeed()) {
	blueSerial.print(".");
      }
      else {
	if (nextUpdate == -2) {		  //  weitere Auslösung einer Reihe?
	  blueSerial.print(" (");
	  blueSerial.print(numFrames - 1);
	  #ifdef TEMPSENS
	    bluePrint_P(PSTR(", "));
	    bluePrintDecimals(ReadTemperatureX10(), 1);
	    bluePrint_P(PSTR("°C"));
	  #endif // TEMPSENS
	  blueSerial.print(")");
	}
	blueSerial.println();
	blueSerial.print((int)(GetSpeed() / 1000));
	blueSerial.print("s: ");
	deltaStern = GetSpeed() >> 4;	//  in 16 Abschnitte unterteilen
      }
    }
    else if (timeReached(nextUpdate, now)) {
      blueSerial.write('*');
    }
    else {				// noch warten
      break;
    }
    //
    // Zeit für nächstes Update = jetzt + 1
    //
    if (IsFastSpeed()) {		// Kamera nicht auf Bulb
      nextUpdate = now + 1000;		//  jede Sekunde
    }
    else {				// lange Zeiten:
      nextUpdate = now + deltaStern;	//  nächster der 16 Abschnitte
    }
    break;
  }
  return shootState == stIdle;		// fertig, wenn 0 erreicht
}

#endif
