/////////////////////////////////////////////////////////////////////////////////////////////
//		 Arduino Intervall-Timer mit Bluetooth-Modul
//
//       17.05.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////

#include "IntervallTimer.h"

#include <EEPROM.h>


//
// Reset/initialise IntervallTimer and load settings from EEPROM if avl. or basic value set.
//
void IntervallTimer::Reset() 
{  
  helpPage = 0;
  calibState = 0;
  sentNewline = 0;
  numDigits = 0;
  inputMode = 0;
  shootState = stIdle;
  numFrames = 0;
  SetShootLine(false);
#ifdef COOLER_PWM
  PID_delay = -1;
  ResetPID(true);
#endif

  ResetSettings(true);	// load settings from EEPROM (if avl)
}

//
// Load settings from EEPROM (if avl && fromEEPROM==true) or basic value set.
//
// Returns true if loaded from EEPROM
//
bool IntervallTimer::ResetSettings(bool fromEEPROM)
{
  if (fromEEPROM && LoadSettingsfromEEPROM()) {
    #if defined(COOLER_PWM)
      SetCoolingLevel(coolingLevel);
    #endif
    #if defined(COOLER_POWER)
      digitalWrite(COOLER_POWER, coolingLevel<0 ? !COOLER_POWER_ON : COOLER_POWER_ON);
    #endif
    #if defined(HEATER_PWM)
      analogWrite(HEATER_PWM, heatingLevel);
    #endif
    #if defined(HEATER_POWER)
      digitalWrite(HEATER_POWER, heatingLevel<0 ? !HEATER_POWER_ON : HEATER_POWER_ON);
    #endif 
    return true;
  }

  // Basis-Einstellungen
  //
  speedIndex = 2;     // 60s
  limitIndex = 2;     // 12h
  quiet = false;
  debug = false;
  bulbBreak_ms = bulbBreak_min;   // Pause zwischen Bulb-Auslösungen, Minimalwert
  speedDither = 0;
  tempCorr = 0;
#if defined(COOLER_PWM) || defined(COOLER_POWER)
  SetCoolingMode(-1);
#endif
#ifdef HEATER_POWER
  SetHeatingMode(-1);
#endif 
  return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Belichtungszeiten in Millisekunden
//
const uint32_t speedTbl[] = {100, 30 sec, 60 sec, 100 sec, 180 sec, 300 sec, 600 sec, 1200 sec, 2400 sec, 3600 sec}; // BeLi-Zeiten in sec
//        --> in Minuten:                  1                 3        5       10        20        40        60 min


// Zeitgrenzen in Millisekunden, 0 == endlos.
//
const uint32_t limitTbl[] = {0, 16*60 Min, 12*60 Min, 8*60 Min, 6*60 Min, 4*60 Min, 2*60 Min, 60 Min, 30 Min, 20 Min, 10 Min}; 


//
// Gewählte Belichtungszeit in der Basiseinheit (ms) zurückgeben
//
uint32_t IntervallTimer::GetSpeed() const
{
  if (speedIndex < 0) {
    return (-speedIndex) sec;	  // negativer Index == Zeit in Minuten
  }
  return speedTbl[speedIndex];
}

//
// Gewählte Zeitgrenze in der Basiseinheit (ms) zurückgeben
//
uint32_t IntervallTimer::GetLimit() const
{
  if (limitIndex < 0) {
    return (-limitIndex) Min;	  // negativer Index == Zeit in Minuten
  }
  return limitTbl[limitIndex];
}

//
// Neue Zeitgrenze in min eintragen und ggf. anwenden, neuen Status anzeigen
//
// Return true if value is ok (1 - MAXINT).
//
bool IntervallTimer::SetLimit(int32_t minutes) 
{
  if (minutes > 0 && minutes <= 0x7FFFL) {	  // nur positive Werte!
    limitIndex = -(int)minutes;	    // in Minuten eintragen
    SetShootLimit();		    // auf laufende Serie anwenden
  }
  else if (minutes == 0) {	    // endlos
    limitIndex = 0;
  }
  else {
    SendErrorBlue();
    return false;
  }
  SendInfoBlue(2);
  return true;
}

//
// Gewählte Zeitgrenze in der Basiseinheit (ms) zurückgeben
//
int IntervallTimer::GetNumSpeeds() const
{
  return (int)ELEMENTS(speedTbl);
}

//
// Gewählte Zeitgrenze in der Basiseinheit (ms) zurückgeben
//
int IntervallTimer::GetNumLimits() const
{
  return (int)ELEMENTS(limitTbl);
}


//////////////////////////////////////////////////////////////////////////
//
// Save settings to EEPROM
// 
// Einstellungen:
//
#define SETTINGS_EEPROM_SIZE  (4 + sizeof(ivtSettings) + sizeof(debug))

void ivtSettings::SaveSettingstoEEPROM()
{
  int iEE = IntervallTimer::GetSettingsEEPROMAddress();	  // EEPROM address
  EEPROM.update(iEE++, 'S');	  // magic
  EEPROM.update(iEE++, 'E');
  EEPROM.update(iEE++, 'T');
  EEPROM.update(iEE++, SETTINGS_EEPROM_SIZE);
  uint8_t* p = (uint8_t*)&debug;
  for (unsigned int i = 0; i < sizeof(debug); i++) {
    EEPROM.update(iEE++, *p++);
  }
  p = (uint8_t*)this;
  for (unsigned int i = 0; i < sizeof(ivtSettings); i++) {
    EEPROM.update(iEE++, *p++);
  }
}


//
// Clear settings from EEPROM
//
void ivtSettings::ClearSettingsfromEEPROM()
{
  EEPROM.update(IntervallTimer::GetSettingsEEPROMAddress(), 0);	  // clear magic only
}


//
// Load settings from EEPROM
// 
// Returns false if not avl.
//
bool ivtSettings::LoadSettingsfromEEPROM()
{
  int iEE = IntervallTimer::GetSettingsEEPROMAddress();	  // EEPROM address
  char c[4];
  EEPROM.get(iEE++, c[0]);		  // check magic (survives scetch uploads)
  EEPROM.get(iEE++, c[1]);
  EEPROM.get(iEE++, c[2]);
  EEPROM.get(iEE++, c[3]);	    // data size
  if (c[0] == 'S' && c[1] == 'E' && c[2] == 'T' && c[3] == SETTINGS_EEPROM_SIZE) {
    uint8_t* p = (uint8_t*)&debug;
    for (unsigned int i = 0; i < sizeof(debug); i++) {
      EEPROM.get(iEE++, *p++);
    }
    p = (uint8_t*)this;
    for (unsigned int i = 0; i < sizeof(ivtSettings); i++) {
      EEPROM.get(iEE++, *p++);
    }
    return true;
  }
  return false;
}



#if defined(COOLER_PWM)
//
// Save PID coeffs to EEPROM
//
void IntervallTimer::SavePIDtoEEPROM()
{
  int iEE = GetPIDEEPROMAddress();	  // EEPROM address
  EEPROM.update(iEE++, 'P');	  // magic
  EEPROM.update(iEE++, 'I');
  EEPROM.update(iEE++, 'D');
  EEPROM.update(iEE++, sizeof(KpKiKd));
  uint8_t* p = (uint8_t*)KpKiKd;
  for (unsigned int i = 0; i < sizeof(KpKiKd); i++) {
    EEPROM.update(iEE++, *p++);
  }
}


//
// Clear PID coeffs from EEPROM
//
void IntervallTimer::ClearPIDfromEEPROM()
{
  EEPROM.update(GetPIDEEPROMAddress(), 0);	  // clear magic only
}


//
// Load PID coeffs from EEPROM
// 
// Returns false if not avl.
//
bool IntervallTimer::LoadPIDfromEEPROM()
{
  int iEE = GetPIDEEPROMAddress();	  // EEPROM address
  char c[4];
  EEPROM.get(iEE++, c[0]);		  // check magic (survives scetch uploads)
  EEPROM.get(iEE++, c[1]);
  EEPROM.get(iEE++, c[2]);
  EEPROM.get(iEE++, c[3]);	    // data size
  if (c[0] == 'P' && c[1] == 'I' && c[2] == 'D' && c[3] == sizeof(KpKiKd)) {
    uint8_t* p = (uint8_t*)KpKiKd;
    for (unsigned int i = 0; i < sizeof(KpKiKd); i++) {
      EEPROM.get(iEE++, *p++);
    }
    return true;
  }
  return false;
}

#endif

//////////////////////////////////////////////////////////////////////////
// Fehler über Bluetooth melden
//
void IntervallTimer::SendErrorBlue()
{
  EnterNewLine(PSTR("Ungültige Eingabe!"));
}


//
// Frische Zeile anfangen und s ausgeben.
//
void IntervallTimer::EnterNewLine(PGM_P s)
{
  EnterNewLine();
  bluePrintln_P(s);
}


//
// Frische Zeile anfangen und s ausgeben und neue Zeile anfangen.
//
void IntervallTimer::EnterNewLineLn(PGM_P s)
{
  EnterNewLine();
  bluePrintln_P(s);
  PrintNewLine();
}


//
// Frische Zeile anfangen, wenn das BT-Terminal keine neue Zeile registriert hat.
//
void IntervallTimer::EnterNewLine()
{
  if (sentNewline != 0) {	  // wurde vor (!) dem aktuellen Ereignis LF gesendet?
    sentNewline = 0;		  //  dann verbraucht
  }
  else {
    blueSerial.println();	    // BT-Terminal sendet kein LF: dann neue Zeile
  }
}


//
// Zeilenumbruch ausgeben.
//
void IntervallTimer::PrintNewLine()
{
  blueSerial.println();
  sentNewline = 1;
}


//
// " (" oder ", " anzeigen
//
void IntervallTimer::PrintMore(bool more)
{
  if (more == 0) {
    bluePrint_P(PSTR(" ("));
  }
  else {
    bluePrint_P(PSTR(", "));
  }
}


//
// PMW Stauts anzeigen
//
void IntervallTimer::PrintCoolerStatus()
{
#if defined(COOLER_PWM) || defined(COOLER_POWER)
  bluePrint_P(PSTR("PWM="));
  if (coolingLevel >= 0) {
    blueSerial.print(coolingLevel);
  }
  else {
    bluePrint_P(PSTR("---"));
  }
#endif
}


//
// PMW Stauts anzeigen
//
void IntervallTimer::PrintHeaterStatus()
{
#ifdef HEATER_POWER
  bluePrint_P(PSTR("Heizung "));
  if (heatingLevel > 0) {	      // an:
    #if HEATER_POWER_BY_COOLER
    if (coolingLevel < 0) {
      bluePrint_P(PSTR("..."));	      //  auto-off
    }
    else 
    #endif
    {
    #if defined(HEATER_SOFT_PWM)
      // zur linearen Anzeige müsste der mode gespeichert werden oder sqrt(heatingLevel):
      blueSerial.print(((heatingLevel+1)*100)>>8);
      bluePrint_P(PSTR("%"));
    #else
      bluePrint_P(PSTR("an"));
    #endif
    }
  }
  else {
    bluePrint_P(PSTR("aus"));
  }
#endif
}


//
// Speed anzeigen
//
void IntervallTimer::PrintSpeed()
{
  if (IsFastSpeed()) {
    bluePrint_P(PSTR("Schnelle Auslösungen für kameraseitig gewählte Zeit"));
  }
  else {
    bluePrint_P(PSTR("Bulb-Belichtungszeit: "));
    blueSerial.print((int)(GetSpeed() / 1000));
    bluePrint_P(PSTR("s"));
  }
}


//
// Info über Bluetooth senden, which = Bitmaske:
// -1: alle
// 1: Speed
// 2: Serien-Limit
//
void IntervallTimer::ShowSettings()
{
  EnterNewLine();		  // frische Zeile
  PrintSpeed();
  bluePrint_P(PSTR(" +/- "));
  blueSerial.print(speedDither);
  bluePrint_P(PSTR("%"));
  bluePrint_P(PSTR(" + "));
  int secs = (int)(bulbBreak_ms / 1000);	//  Sekunden
  blueSerial.print(secs);
  int msecs = (int)(bulbBreak_ms - (int32_t)secs * 1000);	// restliche millis
  if (msecs >= 100) {			//  + ggf. Zehntel
    bluePrint_P(PSTR("."));
    blueSerial.print(msecs/100);
  }
  bluePrint_P(PSTR("s Pause"));
  SendInfoBlue(2);	  // Limit

  #if defined(COOLER_PWM)
  if (IsPIDActive()) {
    bluePrint_P(PSTR("Temperaturregelung, Ziel-Temperatur = "));
    bluePrintDecimals(PID_SetTemp, 1);
    bluePrint_P(PSTR("°C"));
  }
  else
  #endif
  #if defined(COOLER_PWM) || defined(COOLER_POWER)
    PrintCoolerStatus();
  #endif
  #ifdef HEATER_POWER
    bluePrint_P(PSTR(", "));
    PrintHeaterStatus();
    bluePrintln_P(PSTR(""));
  #endif

  bluePrint_P(PSTR("Updates "));
  if (quiet) {
    bluePrint_P(PSTR("aus"));
  }
  else {
    bluePrint_P(PSTR("aktiv"));
  }
  bluePrint_P(PSTR(", Debug-Infos a"));
  if (debug) {
    bluePrintln_P(PSTR("n"));
  }
  else {
    bluePrintln_P(PSTR("us"));
  }
  #if defined(TEMPSENS)
    bluePrint_P(PSTR("Temperatur-Korrektur: "));
    bluePrintDecimals(tempCorr, 1);
    bluePrintln_P(PSTR("°C"));
  #endif
}

//
// Info über Bluetooth senden, which = Bitmaske:
// -1: alle
// 1: Speed
// 2: Serien-Limit
//
void IntervallTimer::SendInfoBlue(int which)
{
  EnterNewLine();		  // frische Zeile
  if (which & 1) {
    PrintSpeed();
    if (!IsFastSpeed()) {
      if (speedDither != 0) {
	bluePrint_P(PSTR(" +/- "));
	blueSerial.print(speedDither);
	bluePrint_P(PSTR("%"));
      }
      if (bulbBreak_ms > bulbBreak_min) {	// mit zusätzlicher Pause?
	bluePrint_P(PSTR(" + "));
	int secs = (int)(bulbBreak_ms / 1000);	//  Sekunden
	blueSerial.print(secs);
	int msecs = (int)(bulbBreak_ms - (int32_t)secs * 1000);	// restliche millis
	if (msecs >= 100) {			//  + ggf. Zehntel
	  bluePrint_P(PSTR("."));
	  blueSerial.print(msecs/100);
	}
	bluePrint_P(PSTR("s Pause"));
      }
    }
    int more = 0;
    #ifdef TEMPSENS
      if (IsFastSpeed()) {
        blueSerial.println();
      }
      bluePrint_P(PSTR(" ("));
      /* 10 mal mitteln: Nur für Test, dauert zu lange!
      int meanResponse = ReadTemperatureX10();      
      int meanCnt = 0;
      while (meanCnt < 9) {
	meanCnt++;
	meanResponse += ReadTemperatureX10();      // take the mean of 10 10x responses
      }
      bluePrintDecimals(meanResponse, 2);
      */
      bluePrintDecimals(ReadTemperatureX10(), 1);
      #if defined(COOLER_PWM)
      if (IsPIDActive()) {
	bluePrint_P(PSTR("->"));
	bluePrintDecimals(PID_SetTemp, 1);
      }
      #endif
      bluePrint_P(PSTR("°C"));
      more = 1;
    #endif // TEMPSENS
    #if defined(COOLER_PWM) || defined(COOLER_POWER)
      PrintMore(more);
      PrintCoolerStatus();
      more = 1;
    #endif
    #ifdef HEATER_POWER
      PrintMore(more);
      PrintHeaterStatus();
      more = 1;
    #endif
    if (more != 0) {
      bluePrint_P(PSTR(")"));
    }
    blueSerial.println();
  }
  if (which & 2) {
    if (GetLimit() == 0) {
      bluePrintln_P(PSTR("(Kein Limit)"));
    }
    else {
      bluePrint_P(PSTR("Ende nach "));
      int hrs = (int)(GetLimit() / 3600000);
      blueSerial.print(hrs);
      bluePrint_P(PSTR("h "));
      int mins = (int)(GetLimit() / 60000) - hrs * 60;
      if (mins > 0) {
	blueSerial.print(mins);
	bluePrint_P(PSTR("min "));
      }
      if (!IsFastSpeed()) {
	bluePrint_P(PSTR("bzw. "));
	blueSerial.print((int)(GetLimit() / GetSpeed()));
	if ((which & 1) == 0) {		  // falls nur Limit: 
	  bluePrint_P(PSTR(" Frames a "));	  //  auch Frame-Gesamtzahl anzeigen
	  blueSerial.print((int)(GetSpeed() / 1000));
	  bluePrintln_P(PSTR("s"));
	}
	else {
	  bluePrintln_P(PSTR(" Frames"));
	}
      }
      else {
	blueSerial.println();
      }
    }
  }
  if (which < 0) {	    // aktuellen Status
    bluePrint_P(PSTR("Status: "));
    if (shootState == stShootBulb) {	// Bulb-Auslösung
      bluePrint_P(PSTR("Bulb ("));
      blueSerial.print((int)((millis() - endShoot)/1000));
      bluePrint_P(PSTR("s, ...)"));
    }
    else if (shootState == stIdle) {	// tut nix
      bluePrint_P(PSTR("Bereit"));
    }
    else {				// Serie
      if (!IsFastSpeed()) {
	//bluePrint_P(PSTR("lfd. Serie, "));
	blueSerial.print(numFrames-1);  // Frames bisher
	if (shootLimit != 0) {
	  bluePrint_P(PSTR(" von "));
	  // Gesamt = T/t, wobei t abweichen kann von früheren Werten
	  // Also nicht T/t - n!
	  //
	  long rest = shootLimit - millis();  // Restzeit
	  if (rest <= 0) {
	    blueSerial.print(numFrames);	// letzter läuft
	  }
	  else {
	    blueSerial.print((int)(rest / GetSpeed()) + (numFrames - 1));
	  }
	}
	bluePrint_P(PSTR(" Frames, + "));
	if (shootState == stShootPaused) {
          blueSerial.print((int)(GetSpeed()/1000));
	}
	else {
          blueSerial.print((int)((GetSpeed() - (endShoot - millis()))/1000));
	}
	bluePrint_P(PSTR(" von "));
        blueSerial.print((int)(GetSpeed()/1000));
	bluePrint_P(PSTR("s"));
      }
      else {
        bluePrint_P(PSTR("Schnelle Auslösung .."));
      }
    }
    bool xtra = false;
    if (quiet) {
      bluePrint_P(PSTR(" (keine Updates"));
      xtra = true;
    }
    if (debug) {
      if (xtra) {
	bluePrint_P(PSTR(", "));
      }
      else {
	bluePrint_P(PSTR(" ("));
      }
      bluePrint_P(PSTR("Debug-Infos an"));
      xtra = true;
    }
    if (xtra) {
      bluePrintln_P(PSTR(")"));
    }
    else {
      PrintNewLine();
    }
  }
  // Speedtest:
  //blueSerial.println((int)(millis()-mm)); // --> 91 ms für ca. 80 Zeichen, kein Puffer, blockiert alles!!!
  // 2 ms mit BufferedSoftwareSerial!
}

//
// Hilfe über Bluetooth senden.
// Um den Sendepuffer klein zu halten, wird die Hilfe in mehrere
//  Teile zerlegt, die jeweils erst gesendet werden, wenn der
//  Sendepuffer wieder geleert ist.
// 'what' = 1...n zeigt an, welcher Teil ansteht.
/*
Der Intervalltimer kann die Kamera periodisch auslösen. Für Langzeit-Belichtungen wird an der Kamera der 
Bulb-Mode gewählt und der Intervalltimer bestimmt die Belichtungszeit.
Wenn die Belichtungszeit von der Kamera vorgegeben werden soll, wird der Intervalltimer auf die minimale
Belichtungszeit von 0.1 Sekunden eingestellt und löst dann zehn mal pro Sekunde aus.

Durch die Wahl einer verlängerten Pause zwischen den Auslösungen kann die Belichtungszeit unabhängig von der Verzögerung durch die Bildspeicherung exakt definiert werden.
Voreingestellt ist eine Pause von 0.2 Sekunden, die aber von der Kamera bei Bedarf verlängert wird - auf Kosten der nächsten Belichtung. Bei zu kurzen Pausen startet u.U. nur jede zweite Auslösung!
Mit langen Pausen können auch Timelapse-Reihen aufgenommen werden. Dazu muss für den Timer die Belichtungszeit auf mindestens 1 Sekunde gestellt werden.
 Wenn die Kamera die Belichtungszeit vorgibt, können Timelapse-Reihen auch einfach durch lange Belichtungszeit erzeugt werden. (Die Kamera löst erst wieder aus, nachdem der Timer die letzte Belichtung beendet hat.)

...
Die Wahl der Betriebsparameter erfolgt über die zuvor hergestellte Bluetooth-Verbindung durch Eingabe einzelner Befehlszeichen, die sofort ausgeführt werden.
Einige Befehle erwarten eine Ziffernfolge, die vor dem Befehlszeichen eingegeben wird.
Der Intervalltimer verliert seine Einstellungen, wenn die Kamera ausgeschaltet wird.
...
Bei bestehender Bluetooth-Verbindung wird der aktuelle Belichtungsstatus kontinuierlich angezeigt.
*/
void IntervallTimer::ShowHelpOn_s()	{bluePrintln_P(PSTR(" 123s  : Belichtungszeit in Sekunden"));}
void IntervallTimer::ShowHelpOn_m()	{bluePrintln_P(PSTR(" 123m : Belichtungszeit in Minuten"));}
void IntervallTimer::ShowHelpOn_l()	{bluePrintln_P(PSTR(" 123L : Limit für Seriendauer in Minuten, 0 schaltet Limit ab."));}
void IntervallTimer::ShowHelpOn_p()	{bluePrintln_P(PSTR(" 123P : Pause zwischen Frames in 1/10 Sekunden"));}
void IntervallTimer::ShowHelpOn_n()	{bluePrintln_P(PSTR(" 123N : Limit für Seriendauer in Frames, 0 schaltet Limit ab."));}
void IntervallTimer::ShowHelpOn_pcnt()  {bluePrintln_P(PSTR(" 12%   : Belichtungszeit um bis zu +/- Prozent zufällig variieren"));}
void IntervallTimer::ShowHelpOn_stat()  {bluePrintln_P(PSTR(" ?  : Zeige aktuellen Status"));}
//void IntervallTimer::ShowHelpOn_()	{}

// Hilfe zu Heizung 
#if defined(HEATER_POWER)
void IntervallTimer::ShowHelpOn_h()
{
# if defined(HEATER_SOFT_PWM)
  bluePrintln_P(PSTR(" 2nK : Fenster-Heizung, 20=aus, 21..25=Stärke 1 bis 5"));
# else
  bluePrintln_P(PSTR(" 2nK : Fenster-Heizung, 20=aus, 21=an"));
# endif
}
#endif

// Hilfe zu Kühlung + Heizung ohne PID
#if !defined(COOLER_PWM) && defined(COOLER_POWER)
void IntervallTimer::ShowHelpOn_kh()
{
  bluePrintln_P(PSTR(" nK   : Kühlung, 0=aus, 1=an"));
  ShowHelpOn_h();
}
#endif

int IntervallTimer::SendHelp(int what)
{
  switch (what) {
    case 1:
      EnterNewLine();		  // frische Zeile
      bluePrintln_P(PSTR(" Intervalltimer Befehle:"));
      bluePrintln_P(PSTR(" !  : Starte/Beende Belichtungssequenz"));
      bluePrintln_P(PSTR("      (zeigt 15 Teilintervalle an)"));
      bluePrintln_P(PSTR(" # : Starte/Beende Dauer-Belichtung (nur im"));
    break;
    case 2:
      bluePrintln_P(PSTR("       Bulb-Mode, zeigt abgelaufene Sekunden an)"));
      bluePrintln_P(PSTR(" + : Belichtungszeit erhöhen"));
      bluePrintln_P(PSTR("  - : Belichtungszeit verringern"));
      ShowHelpOn_s();
    break;
    case 3:
      ShowHelpOn_m();
      bluePrintln_P(PSTR(" > : Erhöhe Limit für Seriendauer"));
      bluePrintln_P(PSTR(" < : Verringere Limit für Seriendauer"));
      ShowHelpOn_l();
    break;
    case 4:
      ShowHelpOn_n();
      ShowHelpOn_p();
      ShowHelpOn_pcnt();
    break;
    case 5:
      ShowHelpOn_stat();
#if defined(COOLER_PWM) || defined(COOLER_POWER) || defined(HEATER_POWER)
# if defined(COOLER_PWM)
      bluePrintln_P(PSTR(" K : Kühlung, Hilfe"));
# else // nur COOLER_POWER:
      ShowHelpOn_kh();
# endif
#endif
      bluePrintln_P(PSTR(" E : Einstellungen, Hilfe"));
    break;
    case 6:
    break;
    case 7:
      bluePrintln_P(PSTR(" H : Zeige diese Hilfe (h=H, r=R, q=Q, ...)"));
      bluePrintln_P(PSTR(" R : Reset (= 2E, lade Einstellungen)"));
      bluePrintln_P(PSTR(" U : Keine Updates senden"));
      bluePrintln_P(PSTR(" D : Debug-Nachrichten senden"));
    return 0;	  		  // fertig mit Basis-Hilfe
    case settingsHelpPage:
      bluePrintln_P(PSTR(" Einstellungen:"));
      ShowHelpOn_stat();
      bluePrintln_P(PSTR(" 0E : Zeige aktuelle Einstellungen"));
      bluePrintln_P(PSTR(" 1E : Lade Grundeinstellungen"));
    break;
    case settingsHelpPage+1:
      bluePrintln_P(PSTR(" 2E : Lade gespeicherte Einstellungen"));
      bluePrintln_P(PSTR(" 3E : Speichere aktuelle Einstellungen im EEPROM"));
      bluePrintln_P(PSTR(" 4E : Lösche gespeicherte Einstellungen"));
    return 0;	  		  // fertig mit Einst.-Hilfe

    case calibHelpPage:
#if defined(COOLER_PWM) || defined(COOLER_POWER)
      bluePrintln_P(PSTR(" Kühlung:"));
# if defined(COOLER_PWM)
#  if defined(COOLER_POWER)
      bluePrintln_P(PSTR(" nK   : 0=aus, 1=minimal, 2=niedrig, 3=std, 4=hoch, 5=maximal,"));
#  else
      bluePrintln_P(PSTR(" nK   : 0,1=minimal, 2=niedrig, 3=std, 4=hoch, 5=maximal,"));
#  endif
      bluePrintln_P(PSTR("         7: Ziel-Temperatur eingeben, Regelung starten"));
      bluePrintln_P(PSTR("         8: Regelung starten, Temperatur halten"));
# endif
#endif
    break;
    case calibHelpPage+1:
# if defined(COOLER_PWM)
      bluePrintln_P(PSTR("         9: Regelung in 10 min starten + halten"));
      bluePrintln_P(PSTR(" 1nnnK : PWM-Signal = nnn"));
      ShowHelpOn_h();
#endif
    break;
    case calibHelpPage+2:
      #ifdef COOLER_PWM
      bluePrintln_P(PSTR(" 0C : PID-Koeffizienten anzeigen"));
      bluePrintln_P(PSTR(" 1C : ... auf Basiseinstellung"));
      bluePrintln_P(PSTR(" 2C : ... aus EEPROM laden"));
      bluePrintln_P(PSTR(" 3C : ... in EEPROM speichern"));
      bluePrintln_P(PSTR(" 4C : ... aus EEPROM löschen"));
      bluePrintln_P(PSTR(" 5C : ... eingeben"));
      #endif
    break;
    case calibHelpPage+3:
      #ifdef COOLER_PWM
      bluePrintln_P(PSTR(" nnC : Kalibriere Kühlung: nn = PWM-Sprung"));
      bluePrintln_P(PSTR("     C : Kalibrierung abbrechen / beenden"));
      #endif
      #ifdef TEMPSENS
      bluePrintln_P(PSTR(" T : Temperaturkorrektur eingeben"));
      #endif
    return 0;		  // fertig mit Kalib.-Hilfe

    default:
    return 0;		  // ?
  }
  return what + 1;		  // weiter, wenn Sendepuffer geleert
}

