///////////////////////////////////////////////////////////////////////////
// Canon EOS Shutter-Emulation on Arduino (Nano / Pro Mini, / ...)
//  for 6D, 6Dii, 50D, 60D, 70D, 80D, 550D+600D, 
//      650D+700D, 1100D, 1200D+1300D, 2000D,
//      EOS M, M3, M10, M100 (details see blow).
//
//			      Version 4.0
//
//       25.1.2017 - 12.09.2024                       Erwin Lotter
//
// 
// The basic Shutter-Emulation can be complemeted (or replaced) by 
// 
//  - an astro shooting module that implements an intervalometer and
//  - a temperature control / cooling module which includes a
//    PID controller,
// 
// which both are operated via a Bluetooth module attached to the arduino.
// 
// The desired EOS or EOSM model is selected in the 'camera.h' file.
// The assignment of the Arduino pin functions to the shutter pins
//  is also defined there.
// Additional defines can be added to configure shooting and/or cooling.
// (See EOS2000D + MY_2000D_2 for a fully featured example.)
// Arduino Pin Functions
//
#include "camera.h"             // select camera model
//
// The file 'cooling.h' holds the definitions and code needed to
// configure a temperature sensor and PID controller(s).
// The required Bluetooth code is part of the intervalometer module,
// which needs to be included below.
// 

//
// In 'pins.h' the connection between Arduino and Bluetooth module
// is configured. Predefined is the use of two unused data lines.
// If
//    //#define BLUESERIAL_IN_SOFTWARE 0  
// in 'pins.h' is uncommented, the serial port of the Arduino is
// used instead. (In this case, Tx of the Bluetooth module must
// not be plugged in during uploads!)
// 

//#define SERIAL_DEBUG  57600	// undef oder baudrate (max 57600!). 
			        // gets undef, if Bluetooth via 'Serial' is choosen 
//#define DEBUG_SLOW    // ToggleDebugLED() aktivieren, evtl. 64x langsamer
//#define DEBUG_BLINK	// BlinkLED() aktivieren
#define BLUE_DEBUG	// Debug-Meldungen über Bluetooth (wenn verfügbar)


//
// If desired, the intervalometer module with Bluetooth communication
//  can be integrated now.
// As a minimum, the files 'pins.h', 'IntervallTimer.h', 'IntervallTimer.ino'
//  and 'Bluetooth.ino'  must be included in the sketch directory. 
// If temperature control is configured in 'camera.h', the files
//  'cooling.h' and 'cooling.ino' and for PID control 'StepResponse.h'
//  are also needed.
//
// 'pins.h' defines the shooting (SHOOT_LINE) and the Bluetooth (blueRX, blueTX)
//  pins of the Arduino. Make sure that these pin definitions are compatible
//  with your hardware (see comments in 'pins.h').
// 
//* // Uncomment to include intervalometer module & Bluetooth
//
#include "IntervallTimer.h"

IntervallTimer ivt;               // Intervall-Timer mit Bluetooth-Anbindung

#define INTERVALL_TIMER ivt
//*/


//
// The Canon EOS shutter-emulation below is inspired by the emulator for the
//  EOS 1100D presented by Hendrik Beijeman (http://www.phym.nl/wp/?p=367).
//
// This sketch should be uplaoded with a programmer and replace the Arduino
//  boot code in order to reduce the boot delay. 
// For testing, the sketch can be uploaded as a sketch only. Then, the arduino
//  power supply voltage must be supplied independently at least one second before
//  the camera is powered up. Otherwise, the camera may hang on startup!
// 
// *** Alternatively, channels which need to be low on startup can be pulled down by
//  a 2k resistor, which takes this job until the Arduino output gets active.
//  With this, it is sufficient to run the emulation code as a simple sketch. ***
//


/****************************************************************************************************
// Aktueller Stand:

  Achtung: Die Emulation ist nur für bestimmte Einstellungen der Kameras geeignet!
	   Die Spiegel-Vorauslösung sollte (sofern verfügbar) abgeschaltet und für den 
	   Live-View der Modus "Leise 1" gewählt sein (Ausnahme: 60D, Leise-Modus aus).
      Während des Einschaltens sollte nicht ausgelöst werden, z.B. durch eine arretierte Fernauslösung.

  6D:
    - Shutter und Spiegel werden emuliert (mit Coil 1 und 2).
    - Auslösen klappt bis 1/2500s, im LV bis 1/1600s, darunter tritt Err30 auf.
    - Auslösen im LV klappt nur im Leise-Modus 1 (Leise-2 hängt, Normal -> Err30). Lässt sich
      evtl. wie bei der 60D dadurch beheben, dass ROT-Signale verzögert werden.
    - Spiegelvorauslösung führt zu Error 20.
    - Abschalten während einer Belichtung führt zu Error 20. 
    - Err20 nach Rückkehr aus Standby?
    - Kleine Blendenöffnungen (Abblenden um mehr als xx Stufen, ab Blendenwert 16 und höher bei einem EF 35-80, ab 5.6 beim EF50 1/1.8)
	    führen beim Auslösen zu Error 30. Dabei wird noch ein zu kurz belichtetes Bild aufgenommen.
	    Ab Mai 2022 sollte das behoben sein!
    - Gelegentlich (alle paar hundert Aufnahmen) hatte 6D-2 Error 30 (wie bei der 600D?).
      Tritt seit V3.2, die das Blenden-Problem behebt nicht mehr auf.
    - Die 6D-5 hatte nach einiger Zeit beim Einschalten immer öfter Error 20. Wird behoben 
      durch 2k2 Pull-Downs an Arduino D6, D7, D10. Offenbar ist hier ein 8 MHz Board kritisch!

  6Dii:
    - Shutter und Spiegel-Lichtschranke werden emuliert, der Spiegelmotor muss aber angeschlossen bleiben.
    - Auslösen klappt bis 1/2500s, im LV bis 1/1600s, darunter tritt Err30 auf.
    - Kleine Blendenöffnungen sind möglich.
    - Leise, schnelle und Serien-Auslösungen werden (noch) nicht unterstützt.
    - Auslösen im LV geht schnell (< 0.5s Verlust) und ohne Spiegelbewegung!
    - Auslösen außerhalb des LV benötigt Belichtungsmessung auch im M- und Bulb-Mode!!!

  50D: Ok, auch im LV und mit Spiegelvorauslösung. Shoot im Leise-Modus 1 und 2 für
       Zeiten >= 1s (oder Bulb, egal wie lang) erzeugt Error 20.

  60D: Mit Coil2 (Leise-Modus muss AUS sein? Oder doch Leise1?)

  70D: Mit Coil2, Leise-Modus 1.
    - Auslösen klappt bis t >= 1/2000s auch im LV, darunter tritt Err30/Err20 auf.
    - Abschalten während einer Belichtung führt zu Error 20. (Modus-Änderung beendet ML Bulb-Timer korrekt!)
  
  80D: Shutter ok für t >= 1/2500, auch im LV-Leise1 aber nicht im Leise-Aufnahmen-Modus.
       Bei sehr kleinen Blenden tritt bei LV-Start Err30 auf. Vmtl. wie bei der 6D zu beheben, s. "80D_BLENDE"!
       Spiegelemulation gibt Err20 bei Shoot und beim Beenden des LV, Shoot im LV ist aber ok.
       Deshalb muss der Spiegelmotor mit seinen Lichtschranken angeschlossen bleiben.
       Wie bei der 6Dii wäre wohl auch eine Emulation der Spiegel-Lichtschranken möglich,
        was deren Verschleiß vermeiden würde.
    - Auslösen im LV geht schnell (< 0.5s Verlust) und ohne Spiegelbewegung!
    - Auslösen außerhalb des LV benötigt Belichtungsmessung auch im M- und Bulb-Mode!!!

  200D: Nur Entwurf! LV-Verwendung vmtl. nicht möglich!

  600D:
    - ohne Coil 2 ist fast alles ok.
    - mit Coil 2 geht alles - etwas ist aber unklar (s. "??600D").
      ACHTUNG: In der eo63a (nur Coil 1) lassen sich keine Zeiten < 1/200 realisieren: Kamera begrenzt!
      ACHTUNG: Das c2_H Timing des Main -13 passt nicht zu den früher gemessenen Werten.
      Ist vermutlich Folge der emulierten ROT- und OPT-Signale!
    - Kleine Blendenöffnungen (Abblenden auf > F11 beim EF50 1/1.8) führten für
      die eo63a (nur Coil 1) beim Auslösen zu Error 30, auch im Bulb Mode - aber nicht im LV.
      Anders als bei der 6D wird aber keine falsche Aktion gewählt, die Kamera
       erwartet scheinbar ein zusätzliches Signal. 
      Ist jetzt behoben durch ein zusätzliches OPT2 Lo,Hi in c1_L_Start. 
      (Unklar ob das Problem auch mit Coil 2 auftritt.)
    - Gelegentlich (alle paar hundert Bilder) Error 30 (bei der 600D-3?).
      Bei der 6D-2 wurde das vmtl. durch das Blenden-Problem verursacht, s. dort "Blendenwerte"
       
  700D:
    -  mit Coil 2: Die Zeiten 1/4000 und 1/3200 ergeben Err 30, Rest ist ok
    - ohne Coil 2: Auslösen im LV klappt erst für t > 1/60s.

  1100D: Außerhalb des LV Err30 für t <= 1/500s! (Shoot im LV klappt auch in der alten Version).
  1200D: Außerhalb des LV Err30 für t <= 1/200s! (Shoot im LV klappt auch in der alten Version).
  1300D: Vmtl. wie 1200D oder 2000D.
  2000D: Ok in allen Betriebsarten.

  M, M3, M50, M100: Ok.
  M10: Vmtl. identisch mit der M100.
*/

/////////////////////////////////////////////////////////////////////////
#if !defined(SERIAL_DEBUG) && !defined(INTERVALL_TIMER)
#define ALLOW_SLEEP		  // Deep Sleep nur wenn nicht kommuniziert wird
#endif

#include "debugmsg.h"		  // erst nach IntervallTimer und Def. von SERIAL_DEBUG!

#include "digitalWriteFast.h"


// Mögliche Ergebnisse eines Trigger-Ereignisses (Interrupts)
//
#define evNone       0
#define evCoil1Start 1
#define evCoil1Stop  2
#define evMotorOn    3
#define evCoil2Start 5 
#define evCoil2Stop  6


struct sAction {
  unsigned long time;
  unsigned int  pin;
  unsigned int  value;
};

#define ms	*1000L       		// Zeit in Millisekunden in die Basiseinheit (derzeit µs) umrechnen. Achtung: ggf. Klammern!

// Werte für sAction::value (zusätzlich zu HIGH und LOW)
#define STOP_VAL	2
#define HALT_VAL	3
#define LONGWAIT_VAL	4
#define BLINK_VAL	5
#define MOTOR_VAL	6
#define STOP_ACTION	{0, 0, STOP_VAL}    // ACHTUNG: Mit STOP_ACTION wird der Sleep-Mode gestartet, in dem die Zeitzählung anhält!!!
#define HALT_ACTION	{0, 0, HALT_VAL}
#define LONGWAIT_ACTION	{0, 0, LONGWAIT_VAL}
#define BLINK_ACTION	{0, 0, BLINK_VAL}   // toggled LED, wenn DEBUG_BLINK definiert ist: Nicht als 1. Eintrag!
#define MOTOR_ACTION	{0, 0, MOTOR_VAL}   // aktiviert Motor-Interrupt oder Warteschleife

volatile int triggerEvent;    // Flag für Trigger-Flanken
sAction* actionPtr;
unsigned long timingZero;   // Start-Zeit

inline void SetPin(int pin, int val) __attribute__((always_inline));

#include <avr/sleep.h>


///////////////////////////////////////////////////////////////////////////////////////////////
#if defined EOS_MXXX	// switch between EOS and EOSM (loop() in ShutterEmuEOSM.ino)

#if defined(EOS_M)
#include "CamTimings\EOS_M.cpp"
#elif defined(EOS_M3)
#include "CamTimings\EOS_M3.cpp"
#elif defined(EOS_M50)
#include "CamTimings\EOS_M50.cpp"
#elif defined(EOS_M100)
#include "CamTimings\EOS_M100.cpp"
#endif

#else // defined EOS_MXXX
//
// Für Modelle, die nur eine Coil haben (EOS M, 1100D, ...) wird diese Einstellung ignoriert.
// Für Modelle mit 2 Coils, von denen aber nur Coil 1 angeschlossen ist, wird auf der 
//  Basis des Coil 1 Signals emuliert, was aber mit Einschränkungen speziell im 
//  LiveView verbunden ist.
// Die 6D und die 80D erzwingen Coil 2, weil Coil 1 bei EnterLV inaktiv bleibt.
//
#define TRIGGER2_MODE	RISING	  // Verwendung des Coil 2 Signals aktivieren durch Angabe der Triggerflanke, die Coil 2 löst (= RISING).

// Signalpegel im Ruhezustand (für Setup()):
// (Allgemeine Vorgabewerte, die für einzelne Modelle überschrieben werden.)
#define	ROTX_INIT	LOW
#define	ROT1_INIT	LOW	  // ovrrd für 60D
#define	ROT2_INIT	HIGH	  // ovrrd für 60/70/80D
#define	ROT3_INIT	HIGH	  // ovrrd für 60D
#define	ROT4_INIT	HIGH
#define	MR1_INIT	HIGH
#define	MR2_INIT	LOW
#define	MR3_INIT	HIGH	  // ovrrd für 50D
// Die OPT1,2-Startpegel werden ohne Def. als HIGH angenommen


///////////////////////////////////////////////////////////////////////////////////////////////
// Verwendung der Coils als Trigger:
// Trigger 1 ist für alle Modelle zwingend, Trigger 2 muss
//  modellspezifisch mit TRIGGER2_MODE (--> 700D) aktiviert werden.
// Die 6D und 80D erzwingen Coil 2, weil Coil 1 bei EnterLV inaktiv bleibt.
//
#if defined(EOS6D)  ||  defined(EOS80D)
#define TRIGGER1_MODE RISING	  // von Coil 1 nur Stopp-Flanke auswerten
#undef  TRIGGER2_MODE
#define TRIGGER2_MODE CHANGE	  // Coil-Aktivierung an Coil 2 festmachen
#else
#define TRIGGER1_MODE CHANGE
# if defined(EOS1100D) || defined(EOS1200D_1300D) || defined(EOS2000D)
# undef TRIGGER2_MODE		  // nicht vorhanden
# endif
#endif

#define TRIGGER1 COIL1
#define TRIGGER1_ACTIVE LOW
#define TRIGGER2 COIL2
#define TRIGGER2_ACTIVE LOW

//////////////////////////////////////////////////////////////////////////////
inline sAction* SelectFinishAction(unsigned long dt) __attribute__((always_inline));
int prevEvent;
unsigned long triggerTime;		// Trigger-Zeitpunkt (von micros()




#if defined(EOS650D_700D)
#include "CamTimings\EOS650D_700D.cpp"
#elif defined(EOS600D)
#include "CamTimings\EOS600D.cpp"
#elif defined(EOS1100D) || defined(EOS1200D_1300D) || defined(EOS2000D)
#include "CamTimings\EOS1100D_2000D.cpp"
#elif defined(EOS6D)
#include "CamTimings\EOS6D.cpp"
#elif defined(EOS80D)
#include "CamTimings\EOS80D.cpp"
#elif defined(EOS70D)
#include "CamTimings\EOS70D.cpp"
#elif defined(EOS60D)
#include "CamTimings\EOS60D.cpp"
#elif defined(EOS60D_alt)
#include "CamTimings\EOS60D_alt.cpp"
#elif defined(EOS50D)
#include "CamTimings\EOS50D.cpp"
#endif


int stateIndex;      				// Status im Ablauf
inline unsigned long timeElapsed() __attribute__((always_inline));
#define state_Off -1


void loop()
{
  //
  // Ablaufschema:
  // ------------------------------------
  // Vor einer Auslösung oder bei länger geöffnetem Shutter
  //  befindet sich der Prozessor im Standby, aus der ihn eine Flanke von Coil 1 oder
  //  ggf. Coil 2 weckt und in triggerEvent einen Code einträgt.
  //
  // Der Trigger weckt auch den Prozessor, so dass loop() wieder durchlaufen wird,
  //  wo dann die zu triggerEvent passende Aktionsliste gestartet wird.
  //
  // Bis zum Ende der aktuellen Pulsgruppe wird loop() durchlaufen und geprüft,
  //  ob die in actionPtr[stateIndex].time vermerkte Zeit erreicht ist. 
  //  Wenn ja, wird die hinterlegte Aktion ausgeführt und stateIndex erhöht.
  //  LONGWAIT_VAL und HALT_VAL Einträge halten die Ausführung bis zum nächsten
  //  Trigger an, STOP_VAL beendet die laufende Aktionsliste.
  //
  // Ablauf eines Ereignisses mit Coil 1 (6D, 80D: Coil 2) als Trigger:
  // ------------------------------------------------------------
  // 1.) Zuerst wird geprüft, ob Coil 1 eine Änderung erfahren hat (triggerEvent wurde im Int-Handler gesetzt).
  //  a)  1 -> 0   stateIndex wird auf 0 gesetzt, timingZero erhält die aktuelle Zeit.
  //               (Viele Modelle schalten regelmäßig, andere nur bei Half-Shoot, die Coil-Pwr aktiv.
  //		    Dadurch erzeugen reale Coils Signale, die aber mit einem Emulator nicht auftreten, weil die Coils
  //                gegen einen konstanten Pullup arbeitet, die emulierte Coil-Pwr also immer anliegt!).
  //  b)  0 -> 1   Zweiter Teil (Schließen) beginnt, stateIndex wird bis hinter einen eventuellen LONGWAIT
  //               inkrementiert. Noch anstehende Aktionen werden dabei ausgeführt.
  //               Die positive Coil-Flanke setzt auch die Zeitbasis für den Rest der Signale neu.
  //
  // Mögliche Folgen mit Coil 1 allein:
  //  c1_L_Start - c1_H_FinishShoot
  //  c1_L_Start - c1_H_EnterLV
  //  c1_L_StartInLV - c1_H_FinishLVShoot  +  c1_L_StartInLV - c1_H_EnterLV
  //  c1_L_StartInLV - c1_H_LeaveLV
  //
  // Mögliche Folgen mit Coil 2 (pos. Flanke) allein:
  //  c1_L_Start - c2_H_FinishShoot (mit LONGWAIT bis c1_H)
  //  c1_L_Start - c2_H_EnterLV
  //  c1_L_StartInLV - c2_H_FinishLVShoot (mit LONGWAIT bis c1_H) +  c1_L_StartInLV - c2_H_EnterLV
  //  c1_L_StartInLV - c2_H_LeaveLV  (nicht für 600D: da bleibt C2 inaktiv
  //
  // Der Vorteil von c2_H liegt darin, dass es unabhängig von der Belichtungszeit zu
  //  einem festen Zeitpunkt kommt. Wahlweise können die c2_H Timingtabellen, die auf
  //  c1_H warten müssen, ein LONGWAIT enthalten, oder das c1_H Event wählt eine neue
  //	Tabelle.
  //
  if (triggerEvent == evCoil1Start) {        // Start-Flanke (Coil 1 - oder, wenn 6D, Coil 2 - aktiviert):
    prevEvent = triggerEvent;
    triggerEvent = evNone;
    timingZero = triggerTime;		      //    Zeitpunkt festhalten
    stateIndex = 0;
#if defined(EOS6D)  ||  defined(EOS80D)  ||  defined(EOS50D)	
    actionPtr = SelectStartAction();
#else
    //
    // Außerhalb des LV kann:
    //  a) eine normale Auslösung vorliegen, die mit c1_L_Start behandelt wird, oder
    //  b) ein Start des LV, für den c1_L_Start die ROT-Signale zu weit ausführt.
    //  c1_L_Start stoppt nach 48 ms und wartet auf Coil 2 --> Hi. Der C2_H Zeitpunkt
    //   lässt zwischen a) und b) unterscheiden.
    //
    // Im LV war die letzte Aktion entweder
    //  - coilX_FinishLVShoot, dem wieder ein cX_H_EnterLV folgen muss, oder
    //  - cX_H_EnterLV. Danach kommt entweder
    //  	a) LV-Auslösen oder 
    //  	b) LV-Ende. Dafür ist c1_L_Start ungeeignet, deshalb c1_L_StartInLV verwenden:
    //
    if (actionPtr == c1_H_EnterLV	// c1_H_EnterLV kann auch im Coil2-Modus vorkommen, wenn die nicht angeschlossen ist
#if defined(TRIGGER2_MODE)
      || actionPtr == c2_H_EnterLV
#endif
    // Achtung, für Coil 1 müsste hier eigentlich kommen:				??600D
    //|| actionPtr == c1_H_FinishLVShoot
    // aber	das klappt für die 600D nicht (Err20)!!
    // Ohne diese Prüfung wird nach LV-Shoot das jetzt folgende Re-Enter-LV mit c1_L_Start 
    //  statt mit c1_L_StartInLV (erfolgreich) eingeleitet. Dadurch wird dann bei C1_H das
    //  erste c1_H_EnterLV anstatt dem zweiten (dem Re-EnterLV) ausgewählt. Die sind eh gleich.
    // Der Fehler tritt auch auf, wenn c1_L_StartInLV gleich ist wie c1_L_Start. Er besteht
    //  vmtl. darin, dass nach c1_H_FinishLVShoot das C1_H Timing ncht richtig erkannt wird.
    // besser so Cam-spezifisch def.:
    //#define IS_IN_LV 		(actionPtr == cX_H_EnterLV  ||  actionPtr == c1_H_FinishLVShoot)
    // 
    ) {
      actionPtr = c1_L_StartInLV;
//PRINTLN("c1_L_StartInLV");
    }
    else {
      actionPtr = c1_L_Start;
//PRINTLN("c1_L_Start");
    }
#endif
  } // if (triggerEvent == evCoil1Start) -----------------------------------------------
  
  else if (triggerEvent == evCoil1Stop) { // Coil 1 wurde inaktiv (C1_H):
    //
    // Falls eine Belichtung stattfand, sollte aktuell ein bei C2_H gestartetes LONGWAIT 
    //  laufen (entweder in c2_H_FinishShoot oder in c2_H_FinishLVShoot).
    // Das wird jetzt beendet (und die davor noch anstehenden Signale
    //  werden generiert).
    //
#if defined(TRIGGER2_MODE)	// ############# C1_H nach C2_H: Belichtung beendet
    if (prevEvent == evCoil2Stop) {
      if (stateIndex >= 0) {
 #if defined(EOS600D)
	//
	// die 600D generiert bei LeaveLV keinen C2 Puls. Deshalb via C1_H erkennen:
	//
	if (actionPtr == c1_L_StartInLV) {		// kein C2_H (und im LV)?
	  actionPtr = c1_H_LeaveLV;
	  stateIndex = 0;
	}
	else			// die anderen Möglichkeiten wurden schon via C2_H unterschieden!
 #endif //defined(EOS600D)
	do {							// hinter Wartepunkt in der c2_H-Liste gehen:
//PRINT(".");
	  int v = actionPtr[stateIndex].value;
	  if (v == LONGWAIT_VAL) {				// WAIT beenden und weiter
	    stateIndex++;
//PRINTLN("*+t");
	    break;
	  }
	  else if (v == STOP_VAL || v == HALT_VAL) {	// anhalten mit/ohne in Sleep zu gehen: erledigt
	    break;
	  }
	  else {					// noch nicht beim WAIT angelangt? (sollte nicht passieren)
	    SetPin(actionPtr[stateIndex].pin, v);	//  bis zum WAIT abarbeiten
	    stateIndex++;
	  }
	} while (1);
      }
    }
    else	// falls Coil2 nicht angeschlossen ist, kam auch kein C2_H:
#endif //defined(TRIGGER2_MODE)
    {
/*
if (actionPtr == c1_L_Start) {PRINTLN("c1_L_Start->");}
else if (actionPtr == c1_H_EnterLV) {PRINTLN("c1_H_EnterLV->");}
else if (actionPtr == c1_H_LeaveLV) {PRINTLN("c1_H_LeaveLV->");}
else if (actionPtr == c1_H_FinishShoot) {PRINTLN("c1_H_FinishShoot->");}
else if (actionPtr == c1_H_FinishLVShoot) {PRINTLN("c1_H_FinishLVShoot->");}
else {PRINTLN("???->");}
//*/
      unsigned long dt = triggerTime - timingZero;		// Coil-1 Pulsdauer
      if (actionPtr == c1_L_Start) {		// (noch) nicht im LV?
	if (IsInWindow_EnterLV(dt)) {
	  actionPtr = c1_H_EnterLV;		// dt passt zu LV-Start oder Restart -- FALSCH bei best. Belichtungszeiten (600D: < 1/60s!!!)
	  //PRINTLN("c1_H_EnterLV");
	}
	else {
	  actionPtr = c1_H_FinishShoot;	// zu kurz/lang für EnterLV: reguläres (nicht zu kurzes!) Auslösen
	  //PRINTLN("c1_H_FinishShoot");
	}
      }
      else { // im LV:
	      // a) Shoot im LV endet
	      // b) Re-Enter LV endet (folgt immer nach a)
	      // c) Leave LV endet: Kann nur nach EnterLV kommen, dt ist sehr kurz
	      //
	if (IsInWindow_LeaveLV(dt)) {		// LeaveLV hat kürzesten Puls, also zuverlässig zu erkennen!
	  //PRINTLN("c1_H_LeaveLV");
	  actionPtr = c1_H_LeaveLV;
	}
	else if (IsInWindow_ReEnterLV(dt)
	  #ifdef EOS600D
		 &&  actionPtr != c1_L_StartInLV  // für 600D zusätzlich prüfen: kurze Zeiten im LV liegen unter HiLIMIT_C1_LEAVE_LV
	  #endif	  
	) {
	  //PRINTLN("c1_H_(Re)EnterLV");
	  actionPtr = c1_H_EnterLV;
	}
	else {
	  // zu lange für EnterLV: reguläres (nicht zu kurzes!) Auslösen
	  // geht für 600D schief für kurze Belichtungen. Sicher zu unterscheiden, wenn vorheriger actionPtr gespeichert wird: Der war dann nicht
	  //  c1_H_FinishLVShoot, weil danach immer c1_H_EnterLV folgt
	  //PRINTLN("c1_H_FinishLVShoot");
	  actionPtr = c1_H_FinishLVShoot;
	}
      }
      stateIndex = 0;
    }
    timingZero = triggerTime;							// Timing nach dem WAIT ab jetzt
    prevEvent = triggerEvent;
    triggerEvent = evNone;
  } // if (triggerEvent == evCoil1Stop) -----------------------------------------------------
  //
  //
#if defined(TRIGGER2_MODE)		// Auch Coil 2 Signale
  else if (triggerEvent == evCoil2Stop) {        // Coil-2 deaktiviert: Shutter öffnet
    unsigned long dt = triggerTime - timingZero;		// Coil-2 Pulsdauer
    timingZero = triggerTime;				      // neuer Startpunkt
    stateIndex = 0;
    actionPtr = SelectFinishAction(dt);
    prevEvent = triggerEvent;
    triggerEvent = evNone;
  } // if (triggerEvent == evCoil2Stop) ----------------------------------------
# endif	
  // Prüfen, ob das zum aktuellen Zustand gehörige Delay erreicht ist.
  //
  int v;
  if (stateIndex < 0) {
    v = STOP_VAL;
  }
  else {
    v = actionPtr[stateIndex].value;
    if (v == HIGH || v == LOW) {
      if (actionPtr[stateIndex].time <= timeElapsed()) {
	//
	// Status[stateIndex] ausgeben.
	// stateIndex weiter und ggf. auf state_off oder state_LongWait setzen.
	//
	SetPin(actionPtr[stateIndex].pin, v);
	stateIndex++;
	v = actionPtr[stateIndex].value;
	if (v == STOP_VAL) {
	  //BlinkLED(actionPtr[stateIndex-1].time);
	  stateIndex = state_Off;							// bereit für Sleep
	}
	//if (v == HALT_VAL) BlinkLED(3 ms);
      }
    }
  }
  //
  if (v == HALT_VAL) DoBlinkLED();
  //
  // Wenn die Pause oder das Ende erreicht werden, den Prozessor schlafen legen,
  //  andernfalls in der Loop() weitermachen:
  //
  if (stateIndex == state_Off || v == LONGWAIT_VAL) {
    if (DoBlinkLED() && triggerEvent == evNone) {
#if defined (EOS80D_VOLLEMU)
      if (stateIndex == state_Off) {
	// Emulation des Spiegelsignals MR2 (M4), das schon ca. 2 ms vor Coil 2 LOW wird.
	// Die Versorgung für die Spiegel-Lichtschranken liegt schon ca. 5 ms vor
	//  den Coils an und wird hier genutzt, das MR2_H rechtzeitig zu erzeugen.
	//  Behebt aber die Probleme nicht: Scheinbar ist MR2_H unkritisch!
	if (digitalRead(MOTORDRV) == LOW) {
	  if (stime != 0) {
	    stime = 0;
	  }
	}
	else {
	  if (stime == 0) {
	    stime = micros();
	  }
	  else {
	    if (micros() - stime > (2 ms)) {
	      SetPin(MR2, HIGH);
	    }
	  }
	}
	/*
	if (stime != 0  &&  (micros() - stime) > 2000 ms) {
	  stime = 0;
	  //PRINTLN("MotorDrv");
	  SetPin(MR2, HIGH);
	  //SetPin(MR2, !digitalRead(MR2));
	  //SetPin(MR1, !digitalRead(MR1));
	}
	*/
      }
#elif defined(ALLOW_SLEEP)	// sleep verhindert serielle Ausgabe!
      // Im SLEEP_MODE_STANDBY sind alle Timer aus! Deshalb wird HALT_VAL verwendet,
      //  um auf die Flanke der Coil 2 zu warten, weil für die das Timing benötigt wird.
      // Mit dem SLEEP_MODE_IDLE, der etwa 350 µA braucht, könnte auch diese Wartezeit
      //  mit sleep verbracht werden, aber das lohnt absolut nicht.
      //set_sleep_mode(SLEEP_MODE_PWR_DOWN); // den tiefsten Schlaf auswählen PWR_DOWN
      set_sleep_mode(SLEEP_MODE_STANDBY); // nach 6 Takten wieder da, 35 µA, Timer aus!

      sleep_enable(); // sleep mode einschalten
      sleep_mode(); // in den sleep mode gehen
      sleep_disable(); // sleep mode ausschalten nach dem Erwachen

      // --> "C:\User\Astro\Basteln\Arduino\Arduino schlafen legen.htm"
#endif 
    }
  }
#ifdef INTERVALL_TIMER            // Falls ein Intervall-Timer integriert ist:
  while (triggerEvent == evNone) {   // falls kein neues Ereignis zu bearbeiten ist:
    INTERVALL_TIMER.CheckBlue();              // Bluetooth-Eingaben ggf. bearbeiten
    if (triggerEvent != evNone) {break;}
    INTERVALL_TIMER.Advance();                // Auslösung ggf. fortführen
    if (triggerEvent != evNone) {break;}
    INTERVALL_TIMER.SendUpdate();             // ggf. Meldung über Bluetooth senden
    if (triggerEvent != evNone) {break;}
    INTERVALL_TIMER.UpdatePID();              // PID-Regelung
    while (                       // anstehende Zeichen im Bluetooth-Buffer senden:
         triggerEvent == evNone	  //  wenn kein neues Ereignis stattfand
      && INTERVALL_TIMER.SendNextBlue()       //  mindestens ein Zeichen
      && actionPtr == 0           //  aber nicht mehr, wenn sonst noch was zu erledigen ist
    );
    break;
  }
#endif
}

//
// Explizit prüfen, ob die Cam gerade die Signale abgeschaltet hat.
// Nur sinnvoll, wenn ein Shutter eingebaut ist, der die Coil-Signale
//  niederohmig belastet. Wird aktuell für EOS6Dii_MIRROR benötigt.
// 
// Klappt (für die 6Dii) nicht zuverlässig:
// - ca. 1s nach LeaveLV kommt ein C2_L, das durchkommt weil OPwr offenbar wieder
//   aktiviert wird! Einmal Half-Shoot kurz nach LV-Ende behebt das Problem.
//
inline
bool IsIdle()
{
#if defined(EOS6Dii_MIRROR)		// nur Spiegel-Emu mit Hardware-Shutter?
  if (digitalRead(OPWR) == LOW) {	//  dann Coil-Signale nur akzeptieren, wenn Opt-Pwr aktiv ist
    return true;			//  sonst gibts bei Ready -> Idle aktive Flanken
  }
#endif
  return false;
}

//////////////////////////////////////////////////////////////////////////////////////
// Interrupt-Handler für Coil(1)
//
void coil1_interrupt()
{
  if (IsIdle()) {	// Spiegel-Emu mit Hardware-Shutter?
    return;		//  Ready -> Idle macht aktive Flanken
  }
# if TRIGGER1_MODE == CHANGE	
  if (digitalRead(TRIGGER1) == TRIGGER1_ACTIVE) {    // Coil aktiv
    triggerTime = micros();
    triggerEvent = evCoil1Start;     // Flag für aktive Flanke
//PRINT("C1_L ");
  }
  else
# endif
    if (actionPtr != 0			// Power-Up: Bei externer Versorgung bekommen wir das mit!
# if defined(TRIGGER2_MODE) && (defined(EOS650D_700D) || defined(EOS60D) || defined(EOS60D_alt) || defined(EOS70D))
      // Auch Coil 2 Signale:  Dann beachten, dass bei der 60D/70D/650D/700D					
      //  im LV-Stop Zyklus C1_H zeitgleich zu C2_H kommt. Deshalb
      //  dieses C1_H nicht auswerten, um Überschneidungen zu vermeiden.
      &&  triggerEvent == evNone && actionPtr != c2_H_LeaveLV
      //  (Die 600D generiert in LeaveLV kein C2_H!)
# endif
    ) {
      triggerTime = micros();
# if defined(EOS6D)  ||  defined(EOS80D)
      // Beim LeaveLV generieren C1 und C2 identische Pulse von ca. 27 bzw. 21 ms, von denen
      //  einer - C1_H - unterdrückt werden muss.
      // Wenn C1_H zuerst kommt, liegt das letzte Ereignis ca. 27 bzw. 21 ms zurück und StartInLV wurde zuletzt ausgeführt.
      // Wenn C1_H 0..1ms nach C2_H kommt, ist entweder noch triggerEvent==evCoil2Stop oder schon actionPtr==c2_H_LeaveLV.
      // Um zu vermeiden, dass mit kurzen Belichtungen (< 1/1000s) Konflikte auftreten, muss genau geprüft werden.
      //
      unsigned long dt = triggerTime - timingZero;
      if (    (IsInWindow_LeaveLV(dt)  &&  actionPtr == c1_L_StartInLV) // erstes Ereignis nach C1_L?
	  ||  actionPtr == c2_H_LeaveLV	  // C2_H schon erkannt und bearbeitet?
	    // die folgende Bedingung tritt auch bei sehr kurzen Belichtungen auf und
	    //  führt dazu, das bei der 6Dii 1/3600 oder kürzer Err30 gibt!
	    //
	  ||  triggerEvent==evCoil2Stop	  // C2_H erfolgt, aber noch nicht in Arbeit? 
      ) {
	return;
      }
# endif
      triggerEvent = evCoil1Stop;      // Ende
//PRINT("C1_H ");
    }
  //ToggleDebugLED();
}

#if defined(TRIGGER2_MODE)		// Auch Coil 2 Signale
//////////////////////////////////////////
// Interrupt-Handler für Coil 2 RISING
//
void coil2_interrupt()
{
  if (IsIdle()) {	// Spiegel-Emu mit Hardware-Shutter?
    return;		//  Ready -> Idle macht aktive Flanken
  }

#if TRIGGER2_MODE == CHANGE	
  // Für die 6D übernimmt Coil_2 die Start-Aufgabe, weil C1 beim LV-Start inaktiv (HIGH) bleibt!
  //  Um die Ablauflogik nicht weiter zu verkomplizieren, wird ein evCoil1Start Ereignis
  //  ausgelöst, das ggf. auch von einem parallelen C1_H überlagert werden dürfte (das aber
  //  nicht durchgelssen wird). 
  //
  if (digitalRead(TRIGGER2) == TRIGGER2_ACTIVE) {    // Coil aktiv
    triggerTime = micros();
    triggerEvent = evCoil1Start;     // Flag für aktive Flanke (für die 6D: der Coil 1!) setzen
//PRINT("C2_L ");
    return;
  }
# endif
  if (actionPtr == 0) return;					// Power-Up: Bei externer Versorgung bekommen wir das mit!
  triggerEvent = evCoil2Stop;
  triggerTime = micros();
  ToggleDebugLED();
  //BlinkLED(17 ms);
//PRINT("C2_H ");
}
#endif


//
// Die seit dem Start des aktuellen Events vergangene Zeit in der Basiseinheit (aktuell µs) bestimmen.
//
unsigned long timeElapsed()
{
  // ACHTUNG:
  //  ms und µs werden als unsigned long bereitgestellt -- int sind hier nur 16 Bit breit!!!
  //  µs (von micros()) laufen schon nach 71 min über!
  //  Der Timer bleibt in allen Sleep-Modi außer SLEEP_MODE_IDLE stehen!!!!!!!!!!!
  //  Die Auflösung von micros() ist bei 8 Mhz nur 8µs (bzw. 4µs bei 16 MHz).
  //
  // Sinnvoll wäre die Umstellung der Basiszeiteinheit auf ms, weil die Shutter eh nicht genauer sind!
  //
  /*
  // Wenn die clock seit Start einen Überlauf hatte, ist ihr Wert ist jetzt um
  //  MAXULONG+1 (ca. 4295 sec) zu klein. Ebenso wie die Differenz zum Start.
  //  Dadurch ergibt sich bei der Differenzbildung ein Underflow, der das
  //  Problem behebt. Solange also innerhalb eines Zyklus nur ein
  //  Überlauf auftritt, gibt es kein Problem!
  //
  unsigned long t = micros() - timingZero;
  if ((long)t < 0) {				// Überlauf?
	  t -= MAXULONG+1;				// == 't -= 0' !!!!
  }
  return DebugSlowdown(t);
  */
  return DebugSlowdown(micros() - timingZero);
}

#endif // !defined EOS_MXXX	// switch between EOS and EOSM (code in ShutterEmuEOSM.ino)


// Initialisierung beim Start (oder nach Reset)
void setup()
{
#if defined EOS_MXXX	/////////////////////////// switch between EOS and EOSM

  pinMode(COIL1,  INPUT_PULLUP); // COIL1 mit Pullup
  digitalWriteFast(OPT1, OPT1_SUP);     // Ruhezustand der emulierten Pins
  pinMode(OPT1,  OUTPUT);               //  vor der Aktivierung einstellen
  digitalWriteFast(OPT2, OPT2_SUP);
  pinMode(OPT2,  OUTPUT);

# ifndef EOS_M100
  pinMode(MOTOR, INPUT);
# endif

# ifdef EOS_M50
  pinMode(COIL2, INPUT_PULLUP);
  digitalWriteFast(OPTM, OPTM_SUP);
  pinMode(OPTM,  OUTPUT);
  attachInterrupt(digitalPinToInterrupt(COIL2), coil2_interrupt, COIL_TRIGGER);
# else // nicht M50:
//motorSuppress = 0;
#  if defined(EOS_M)
  attachInterrupt(digitalPinToInterrupt(MOTOR), motor_interrupt, MOTOR_TRIGGER);
#  endif
# endif

  attachInterrupt(digitalPinToInterrupt(COIL1),  coil1_interrupt,  COIL_TRIGGER);

#else //////////////////////////////////////////////////////////// not EOS_MXXX:
  stateIndex = state_Off;

  // Ruhezustand der Pins init.
  // (vor pinMode() mit digitalWriteFast(), damit evtl. per Pull-Down erzeugte
  //  Startwerte nicht kurzzeitig geändert werden!)
  //
  digitalWriteFast(OPT1, HIGH);
# ifdef   OPT2
  digitalWriteFast(OPT2, HIGH);
# endif

  digitalWriteFast(ROT1, ROT1_INIT);
  digitalWriteFast(ROT2, ROT2_INIT);
# ifdef   ROT3
  digitalWriteFast(ROT3, ROT3_INIT);
# endif
# ifdef   ROT4
  digitalWriteFast(ROT4, ROT4_INIT);
# endif
# ifdef   ROTX
  digitalWriteFast(ROTX, ROTX_INIT);
# endif
# ifdef   MR1
  digitalWriteFast(MR1, MR1_INIT);
# endif
# ifdef   MR3
  digitalWriteFast(MR3, MR3_INIT);
# endif
# ifdef   MR2
  digitalWriteFast(MR2, MR2_INIT);
# endif

  //
  // Funktion der emulierten Pins (COIL1 mit Pullup, s.o.):
  pinMode(COIL1, INPUT_PULLUP);
  pinMode(OPT1, OUTPUT);
  pinMode(ROT1, OUTPUT);
  pinMode(ROT2, OUTPUT);
#ifdef   OPT2
  pinMode(OPT2, OUTPUT);
#endif
#ifdef   COIL2
  pinMode(COIL2, INPUT_PULLUP);
#endif
#ifdef   MOTORDRV
  pinMode(MOTORDRV, INPUT);   // 80D
stime = micros();
#endif

#ifdef   ROT3
  pinMode(ROT3, OUTPUT);
#endif
#ifdef   ROT4
  pinMode(ROT4, OUTPUT);
#endif
#ifdef   ROTX
  pinMode(ROTX, OUTPUT);
#endif
#ifdef   MR1
  pinMode(MR1, OUTPUT);
#endif
#ifdef   MR2
  pinMode(MR2, OUTPUT);
#endif
#ifdef   MR3
  pinMode(MR3, OUTPUT);
#endif

  while (!IsIdle()  &&  digitalRead(TRIGGER1) == TRIGGER1_ACTIVE) {SetupDebug();}
  attachInterrupt(digitalPinToInterrupt(TRIGGER1), coil1_interrupt, TRIGGER1_MODE);
# ifdef TRIGGER2_MODE
  while (!IsIdle()  &&  digitalRead(TRIGGER2) == TRIGGER2_ACTIVE) {SetupDebug();}
  attachInterrupt(digitalPinToInterrupt(TRIGGER2), coil2_interrupt, TRIGGER2_MODE);
# endif
#endif // !defined EOS_MXXX	// switch between EOS and EOSM (code in ShutterEmuEOSM.ino)
  
  actionPtr = 0;
  triggerEvent = evNone;
#ifdef INTERVALL_TIMER           // Falls ein Intervall-Timer integriert ist:
  INTERVALL_TIMER.Setup();
#endif
  SetupDebug();
}

// Wert val an Pin pin ausgeben
void SetPin(int pin, int val)
{
  digitalWrite(pin, val);
  ToggleDebugLED();
}

