///////////////////////////////////////////////////////////////////////////
// Canon EOS Shutter-Emulation on Arduino ffor EOSM models
//
// For details see ShutterEmuEOS.ino
//
//       25.1.2017 - 3.08.2024                       Erwin Lotter

#if defined EOS_MXXX	// switch between EOSM and EOS (code see main file ShutterEmuEOS.ino)

/*
#if !defined(EOS_M100)
# define MOTOR_ACTION  {0, 0, MOTOR_VAL}   // aktiviert Motor-Interrupt oder Warteschleife
#endif


#if defined(EOS_M)
#include "CamTimings\EOS_M.cpp"
#elif defined(EOS_M3)
#include "CamTimings\EOS_M3.cpp"
#elif defined(EOS_M50)
#include "CamTimings\EOS_M50.cpp"
#elif defined(EOS_M100)
#include "CamTimings\EOS_M100.cpp"
#endif

void motor_interrupt();
void coil1_interrupt();
void coil2_interrupt();


void setup() {
  triggerEvent = evNone;
  actionPtr = 0;

  pinMode(COIL1,  INPUT_PULLUP); // COIL1 mit Pullup
  digitalWriteFast(OPT1, OPT1_SUP);     // Ruhezustand der emulierten Pins
  pinMode(OPT1,  OUTPUT);               //  vor der Aktivierung einstellen
  digitalWriteFast(OPT2, OPT2_SUP);
  pinMode(OPT2,  OUTPUT);

#ifndef EOS_M100
  pinMode(MOTOR, INPUT);
#endif

#ifdef EOS_M50
  pinMode(COIL2, INPUT_PULLUP);
  digitalWriteFast(OPTM, OPTM_SUP);
  pinMode(OPTM,  OUTPUT);
  attachInterrupt(digitalPinToInterrupt(COIL2), coil2_interrupt, COIL_TRIGGER);

#else// nicht M50:

//motorSuppress = 0;
# if defined(EOS_M)
  attachInterrupt(digitalPinToInterrupt(MOTOR), motor_interrupt, MOTOR_TRIGGER);
# endif
#endif

  attachInterrupt(digitalPinToInterrupt(COIL1),  coil1_interrupt,  COIL_TRIGGER);

#ifdef INTERVALL_TIMER                 // Falls ein Intervall-Timer etc. integriert ist:
  INTERVALL_TIMER.Setup();
#endif
  SetupDebug();
}
*/

// Ablaufschema
// -------------
// Außerhalb der Pulsgruppen, die Events zugeordnet werden, befindet sich
//  der Prozessor im Standby, aus dem ihn eine Flanke von COIL1 oder MOTOR
//  per Interrupt weckt.
//
// Bis zum Ende der aktuellen Pulsgruppe wird loop() durchlaufen und geprüft,
//  ob die in actionPtr->time vermerkte Zeit erreicht ist. 
//  Wenn ja, wird die hinterlegte Aktion ausgeführt und actionPtr erhöht.
//
void loop() {
  //
  // Ablauf eines Ereignisses:
  // --------------------------
  // In den Interrupt-Handlern wird in triggerEvent hinterlegt, welches Ereignis auftrat. 
  // Passend dazu wird hier in actionPtr ein Zeiger auf die Signalsequenz hinterlegt.
  // Anschließend wird diese Sequenz abgearbeitet und abschließend der Sleep-Mode aktiviert, aus
  //  dem der Prozessor erst durch den nächsten Interrupt wieder geweckt wird.
  //
  if (triggerEvent != evNone) {        // neuer Interrupt?
    timingZero = micros();             //    Zeitpunkt festhalten
    if (triggerEvent == evCoil1Start) { // Start-Flanke Coil:
      actionPtr = coilStartAction;
      //PRINTLN("c1_L");
    }
    else if (triggerEvent == evCoil1Stop) {        // Stopp-Flanke (Coil inaktiv):
#ifdef EOS_M3
      // Für eine eventuelle Bulb-Auslösung wurde der Motor-Interupt aktiviert.
      // Falls er nicht ausgelöst wurde, muss er noch deaktiviert werden:
      // 
      detachInterrupt(digitalPinToInterrupt(MOTOR));  
      //
      // Im Bulb-Mode der M3 wird der Shutter viel später aufgezogen.
      // Deshalb anderes Timing vorgeben, wenn zuvor ein Motor-Interrupt kam.
      // Das wird daran erkannt, dass die Ausführung auf einer Warte-Anweisung
      //  steht
      //
      if (actionPtr == M_WAIT) {   // warten nach Motor-Interrupt?
        actionPtr = coilStopBulbAction;            //  weiter für Bulb-Abschluss
        //PRINTLN("c1_H_StopBulb");
      }
      else 
#endif
      {
        actionPtr = coilStopAction;
        //PRINTLN("c1_H_Stop");
      }
    }
#ifndef EOS_M100
    else if (triggerEvent == evMotorOn) {        // Motor-Start (vor c1_H oder nach c1_H):
# ifdef EOS_M3
      if (actionPtr == CS_WAIT) {                // zuletzt Motor-Interrupt in ...?
        //(Alternativ könnte hier auch auf (actionPtr != M_WAIT) geprüft werden!)
        actionPtr = coilFinishBulbAction;            //  weiter für Bulb-Abschluss
        //PRINTLN("Motor_H_StopBulb");
      }
      else 
# endif      
      {
        //PRINTLN("Motor_H");
        actionPtr = motorAction;
      }
    }
# ifdef EOS_M50
    else if (triggerEvent == evCoil2Start) {  // Start-Flanke Coil2:
      actionPtr = coil2StartAction;
      //PRINTLN("c2_L");
    }
    else if (triggerEvent == evCoil2Stop) {   // Stopp-Flanke (Coil2 inaktiv):
      actionPtr = coil2StopAction;
      //PRINTLN("c2_H_Stop");
    }
# endif      
#endif      
    triggerEvent = evNone;
  }
  //
  // Wenn das Ende einer Sequenz erreicht wurde, den Prozessor schlafen legen,
  //  andernfalls in der Loop() weitermachen:
  //
  if (actionPtr == 0) {
#ifdef ALLOW_SLEEP
    //set_sleep_mode(SLEEP_MODE_PWR_DOWN); // den tiefsten Schlaf auswählen PWR_DOWN
    set_sleep_mode(SLEEP_MODE_STANDBY); // nach 6 Takten wieder da, 35 µA
    sleep_enable();  // sleep mode einschalten
    sleep_mode();    // in den sleep mode gehen
    sleep_disable(); // sleep mode ausschalten nach dem Erwachen
#endif
  }
  else {                                            // noch zu tun:
    if (actionPtr->time <= DebugSlowdown(micros() - timingZero)) {  // jetzt?
      SetPin(actionPtr->pin, actionPtr->value);     // Zustand [actionPtr] ausgeben.
      actionPtr++;                                  // actionPtr weiter,
#if defined(EOS_M50)                                // M50: 
      if (actionPtr->value == MOTOR_VAL) {          //  jetzt auf den Motor warten?
        if (digitalRead(MOTOR) == HIGH) {           //  weiter, wenn Motor anläuft (nach 340ms im Bulb, 16ms sonst)
          actionPtr = motorAction;
          //PRINTLN("Motor_Start");     // (zeigt Mist, wenn lange gewartet (Bulb))
        }
        else {
          actionPtr--;                              // weiter warten, wenn Motor noch nicht an
          //PRINT(".");
        }
      }
#elif !defined(EOS_M100)
      if (actionPtr->value == MOTOR_VAL) {
        //motorSuppress = 1; // s.u.
        EIFR = 1 << digitalPinToInterrupt(MOTOR);  // clear flag for pending interrupt, http://gammon.com.au/interrupts
        attachInterrupt(digitalPinToInterrupt(MOTOR), motor_interrupt, MOTOR_TRIGGER);
        actionPtr++;                                // actionPtr weiter,
      }
#endif
      if (actionPtr->value == BLINK_VAL) {
        ToggleLED();
        actionPtr++;                              // actionPtr weiter,
      }
      if (actionPtr->value == STOP_VAL) {         //  löschen wenn fertig
        actionPtr = 0;
      }
    }
  }
#ifdef INTERVALL_TIMER            // Falls ein Intervall-Timer integriert ist:
  while (triggerEvent == evNone) {   // falls kein neues Ereignis zu bearbeiten ist:
    INTERVALL_TIMER.CheckBlue();              // Bluetooth-Eingaben ggf. bearbeiten
    if (triggerEvent != evNone) {break;}
    INTERVALL_TIMER.Advance();                // Auslösung ggf. fortführen
    if (triggerEvent != evNone) {break;}
    INTERVALL_TIMER.SendUpdate();             // ggf. Meldung über Bluetooth senden
    if (triggerEvent != evNone) {break;}
    INTERVALL_TIMER.UpdatePID();              // PID-Regelung
    while (                       // anstehende Zeichen im Bluetooth-Buffer senden:
         triggerEvent == evNone	  //  wenn kein neues Ereignis stattfand
      && INTERVALL_TIMER.SendNextBlue()       //  mindestens ein Zeichen
      && actionPtr == 0           //  aber nicht mehr, wenn sonst noch was zu erledigen ist
    );
    break;
  }
#endif
}

inline bool IsActive(int coil) { return digitalRead(coil) == COIL_ACTIVE; }
//////////////////////////////////////////////////////////////////////////////////////
// Interrupt-Handler
//
// a) Coil (Coil1)
//
void coil1_interrupt() {
  if (IsActive(COIL1)) {             // Coil1 wird aktiv:
    triggerEvent = evCoil1Start;     //  Flag für aktive Flanke
  }
  else {                            ////////////// Coil1 wird inaktiv:
    triggerEvent = evCoil1Stop;      // Ende
  }
  ToggleDebugLED();
}

// b) Coil2 (M50)
//
#ifdef COIL2
void coil2_interrupt() 
{
  if (IsActive(COIL2)) {            ////////////// Coil2 wird aktiv:
    triggerEvent = evCoil2Start;    //  Flag für aktive Flanke
  }
  else {                            ////////////// Coil2 wird inaktiv:
    triggerEvent = evCoil2Stop;     // Coil2 Ende
  }
  ToggleDebugLED();
}
#endif

// c) Motor: Triggert auf RISING. 
//  Hier unbesehen die Abschlussaktion anstoßen. Falls gerade ausgelöst wird, folgt
//   noch Coil. Dann wird diese Motor-Sequenz einfach überschrieben.   
// Alternative zur Erkennung der Auslösung: Motor-Sequenz nur nach evCoil1Stop
//  einleiten: if (previousEvent == evCoil1Stop)...
//  Beim Einschalten der Stromversorgung wird allerdings die Coil kurz aktiviert! Das müsste
//   dann wieder in Setup() init. werden: Auslasszähler = 1
//
#if !defined(EOS_M100) && !defined(EOS_M50)
void motor_interrupt() {
#ifdef MOTOR_GATE
  s.o.;  Eleganter wäre, den Interrupt temporär mit detachInterrupt(digitalPinToInterrupt(MOTOR)) stillzulegen
#endif  
#ifdef EOS_M3
  /*
  if (motorSuppress != 0) {
    // Seltsamerwese kommt bei der M3-Emulation sofort nach attachInterrupt ein
    // Interrupt, der unterdrückt werden muss!
    motorSuppress--;
    return;
  }
  */
  detachInterrupt(digitalPinToInterrupt(MOTOR));  // einmal reicht (Motor ist gepulst!)
#endif
  triggerEvent = evMotorOn;     // Flag für Motor-An
  ToggleDebugLED();
}
#endif


#endif // EOS_MXXX
