/////////////////////////////////////////////////////////////////////////////////////////////
//   PID Calibration for Arduino Intervall-Timer mit Bluetooth-Modul
//
//       17.08.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////
#ifndef StepResponse_h_
#define StepResponse_h_


template <int N, int response_scaling>
class StepResponse {
  public:
    void    Init(int initialResponseDelta_xS, int output_step);
    unsigned Update(int response_xS);	// --> response_scaling
    int     CalculateCoeffs();
    bool    Calculate_T_sum(float* KpKiKd);
    bool    Calculate_ZieglerNichols(float* KpKiKd);
    float*  Get_T_sum()			    {return k_tsum;}
    float*  Get_ZieglerNichols()	    {return k_ZN;}
    int	    GetCount() const		    {return ti;}
    int	    GetLastReading_x1() const	    {return (nextResp - deltaR) / response_scaling;}
    int	    GetReading_x1(int i) const	    {return (firstResp + deltaR * i) / response_scaling;}
    int	    GetTime(int i) const	    {return time_periods[i];}
    int	    GetOutputStep() const   	    {return outputStep;}
    uint32_t GetStartMillis() const	    {return t_start;}

  protected:
    int	    time_periods[N];	  // time line in 128 millisecond units (overflow after 6553 s or 1.8 h)
    int	    deltaR;
    int	    firstResp;
    int	    nextResp;
    int	    ti;
    uint32_t t_start;
    int	    meanCnt;
    int	    meanResponse;
    int	    outputStep;
    float   k_tsum[3];
    float   k_ZN[3];
};


//
// Initialise calibration with actual response level and suggested deltaR.
// Expects that response increases with time. So response must be negatet
//  for cooling and must have the same scaling as used in the
//  PID controller (UpdatePID()). To obtain the correct sign for the 
//  PID coefficients, output_step should be negated too, then.
// The sign of output_step should not reflect the direction of the
//  PWM output. This is handled separately by the PID controller.
// 
// initialResponseDelta_xS should ensure that most of the N response levels will be used,
//  i.e. initialResponseDelta_xS ~= 0.8 (response - final_response) / N.
// If more than N response levels will be needed, deltaR will be doubled.
//
template <int N, int response_scaling>
inline void StepResponse<N, response_scaling>::Init(int initialResponseDelta_xS, int output_step)
{
  outputStep = output_step;
  deltaR = initialResponseDelta_xS;
  meanCnt = 0;
  meanResponse = 0;
  ti = -1;
}


//
// Update Calibration if response level has reached next level.
// Expects that response increases with time. Therefore, response must 
//  be negatet for cooling!
//
// The sampled delays in time_periods[i] indicate at which delay
//  a corresponding response R was reached after the start, where
//    R(i) = (i+1) * step.
//
// Returns the number of time units (128 msec) expired since last sample taken
//  or 0 if building mean or if less than 10 samples taken.
// 
// The response_xS is scaled by a factor of 4 for enhanced resolution compared
//  to the response signal used elsewhere!
//
template <int N, int response_scaling>
inline unsigned StepResponse<N, response_scaling>::Update(int response_xS)
{
  meanResponse += response_xS;		      // take the mean of 8 responses
  if (meanCnt < 7) {
    meanCnt++;
    return 0;
  }
  int r = meanResponse >> 3;		      // mean value (*4)
  meanResponse = 0;
  meanCnt = 0;
  if (ti < 0) {				      // first sample?
    firstResp = r;
    t_start = millis();
    nextResp = r + deltaR;		      // next response level to wait for
    ti = 0;				      // [0] will hold the delay of the first response after the initial
    return 0;
  }
  unsigned now = (millis() - t_start) >> 7;   // duration from beginning (in 128 ms units)
  if (r < nextResp) {		      // next response level not reached:
    if (ti < 10) {			      //  too few samples taken:
      if (now > 10000) {
	return now;			      //  10000*(128/1000) = 1280 s = 21 min is too much 
      }
      return 0;				      //  wait some more time (also ensures that ti > 0)
    }
    else {				      // ti >= 10
      return now - time_periods[ti-1];	      //  return time units (128 msec) expired since last sample taken
    }
  }
  nextResp += deltaR;	// next response level to wait for
  if (ti == (N & 0xfffE)) {   // (even) limit reached?
    //
    // time_periods buffer is filled: pack two periods together
    //
    for (ti = 0; ti < N/2; ti++) {
      time_periods[ti] = time_periods[ti*2] + time_periods[ti*2 + 1];  // pack every two consecutive time_periods together
    }
    deltaR *= 2;		// deltaR between response is doubled now
    return -1;
  }
  time_periods[ti] = now;	  // 1 =^ 128 ms
  ti++;
  if (ti < 2) {
    return now;			  // first sample
  }
  return now - time_periods[ti-2];// gap between this and previous sample
}


//
// Calculate PID coeffs from sampled data by I-TAE tuning rule (T-Summen-Regel)
//  and Ziegler–Nichols method.
// 
template <int N, int response_scaling>
inline int StepResponse<N, response_scaling>::CalculateCoeffs()
{
dbgPrint_P(PSTR("\nN="));
dbgPrint(ti);
dbgPrint_P(PSTR(" delta="));
dbgPrintDecimals((deltaR*10)/response_scaling, 1);

  int result = Calculate_T_sum(k_tsum) ? 1 : 0;
  result |= Calculate_ZieglerNichols(k_ZN) ? 2 : 0;

  return result;
}


//
// Calculate PID coeffs from sampled data by I-TAE tuning rule (T-Summen-Regel)
// 
//  https://en.wikipedia.org/wiki/Proportional–integral–derivative_controller
//
template <int N, int response_scaling>
inline bool StepResponse<N, response_scaling>::Calculate_T_sum(float* KpKiKd)
{
  KpKiKd[0] = 0;    // indicates failure
  if (ti < 4) {
    return false;
  }
  int resp_amplitude = (nextResp - deltaR - firstResp);	// total response range x 4
  float invGain =  (outputStep * response_scaling) / (float)resp_amplitude;	// 1/Gain
  //
  // Ks = 1/Gain
  // Kp >= Ks
  // Ki = Kp / Tn
  // Kd = Kp * Tv
  //
  // Integrate area below curve on the left side and above curve onthe right
  //  in order to find point Ts, where left and right area are equal:
  // 
  //  R(i) = (i+1) * step			  response at t(i)
  //  t(i) = time_periods[i]			  delay from start
  //  A(i) = (t(i)-t(i-1)) * step*(i+0.5)	  area below curve
  // sum(0..n) = step * ((t1-t0)*1 + (t2-t1)*2 + ... +(tn-tn_1)*n + tn/2)
  //           = step * (-t0 - t1 - ... + (n+0.5)*tn)
  //  A(i) = (t(i)-t(i-1)) * step*(i - i_last + 0.5)	  area above curve
  //
  //                 Kp = 1/Gain
  // Tn = 2 Ts / 3,  Ki = Kp / Ts * 1.5
  // Tv = Ts / 6,    Kd = Kp * Ts / 6	  (limit Ts to 100s!?)
  //
  float ALft = 0;
  float ARgt = 0;
  int iLft = 1;		// first sample + 1
  int iRgt = ti-1;	// last sample

  while (iLft < iRgt) {	// integrate from left and right side
    if (ALft <= ARgt) {	//  while keeping left and right  areas equal
      ALft += (time_periods[iLft] - time_periods[iLft - 1]) * (iLft*2+1);
      iLft++;
    }
    else {
      ARgt += (time_periods[iRgt] - time_periods[iRgt - 1]) * ((ti - iRgt)*2+1);
      iRgt--;
    }
  }
  // Now iRgt and iLft have reached the point where the areas left below and
  //  right above are equal.
  // (Times are given in 128ms units and will now be converted to seconds.)
  // 
  float tS = (time_periods[iLft] + time_periods[iRgt]) * 0.064; // / 2  / 1000 * 128;
  if (tS > 100) {
    tS = 100;	  // limit Ts to 100s. Otherwise Ki will vanish and Kd will saturate even for smallest possible error!
  }
  dbgPrint_P(PSTR(" tS="));
  dbgPrint(tS);
  dbgPrint_P(PSTR("s, Ks=1/G="));
  dbgPrintDecimals((int)(invGain*1000), 3);
  if (tS == 0) {
    return false;
  }
  dbgPrintln_P(PSTR(""));
  KpKiKd[0] = invGain;
  KpKiKd[1] = invGain / tS * 1.5;
  KpKiKd[2] = invGain * tS * .167;
  return true;
}


//
// Calculate PID coeffs from sampled data by Ziegler–Nichols method
//  --> https://en.wikipedia.org/wiki/Ziegler-Nichols_method
//
template <int N, int response_scaling>
inline bool StepResponse<N, response_scaling>::Calculate_ZieglerNichols(float* KpKiKd)
{
  KpKiKd[0] = 0;    // indicates failure
  int resp_amplitude = nextResp - deltaR - firstResp;	// total response range
  float invGain =  (outputStep * response_scaling) / (float)resp_amplitude;	// 1/Gain

  // find max slope of response curve:
  //  search for min delay between deltas.
  // Start searching only after 1/8 of response range, to
  //  avoid wrong (too early) detection due to noise.
  //
  int imin = 0;	
  unsigned dtmin = 65000;
  for (int i = (ti >> 3) + 1; i < ti; i++) {
    unsigned dt = time_periods[i] - time_periods[i - 1];
    if (dt < dtmin) {
      dtmin = dt;
      imin = i;		// exact value is i - 0.5!
    }
  }
  if (imin < 3) {
    return false;
  }
  float maxSlope = (float)deltaR / dtmin;  // max slope of response curve

  // Starting from the inflection point of the curve at
  //	tI = time_periods[imin], 
  //	RI = firstResp + deltaR * (imin + 1/2),
  // we calculate the intersection points of the tangent with 
  // initial and final response level (tU and tG resp.):
  //	tU = tI - (RI-R_first) / maxSlope
  //	tG = tI + (RI-R_last) / maxSlope
  //
  //
  int tI = time_periods[imin];
  float dRI = deltaR * (imin + 0.5);	  // additional response at incl. point
  float tU = tI - dRI / maxSlope;
  float tG = tI + (resp_amplitude - dRI) / maxSlope;

dbgPrint_P(PSTR(" imin="));
dbgPrint(imin);
dbgPrint_P(PSTR(" dtmin="));
dbgPrint(dtmin);

dbgPrint_P(PSTR("  tI="));
dbgPrint(tI);
dbgPrint_P(PSTR(" dRI="));
dbgPrint((int)dRI);
dbgPrint_P(PSTR(" tU="));
dbgPrint((int)tU);
dbgPrint_P(PSTR(" tG="));
dbgPrint((int)tG);

  //
  // Kp = 1.2 / G * tG/tU
  // Ki = 0.5 Kp / tU
  // Kd = 0.5 Kp * tU
  // (Times are given in 128ms units and need to be converted to seconds.)
  //
  KpKiKd[0] = 1.2 * invGain * tG / tU;
  KpKiKd[1] = 0.5 * KpKiKd[0] / tU * 0.128;
  KpKiKd[2] = 0.5 * KpKiKd[0] * tU / 0.128;
  if (KpKiKd[2] > 100) {
    KpKiKd[2] = 100;	  // limit Kd to 100. Otherwise Kd will saturate even with very small error!
  }

  return true;
}

#endif
