#ifndef cooling_h_
#define cooling_h_

// PWM output levels for the "1K" to "5K" settings (initial mode is "3K").
// Used to set cooling mode 1 (lowest) to 5 (highest) according to commands "0K" .. "5K".
// Applied in reverse order if COOLER_PWM_REVERSE indicates reverse PWM signal.
// Ignored if COOLER_PWM is not defined.
// 
// Adjust appropriate to hardware response:
//
#ifdef COOLER_PWM
const int CoolerPWMLevels[5] = {0, (int)(COOLER_PWM_PRESET * 0.6), (int)(COOLER_PWM_PRESET), (int)(COOLER_PWM_PRESET * 2.3), 255};
#endif

// PWM output levels for the "21K" to "25K" settings (initial mode is "23K").
// Used to set window heating mode 1 (lowest) to 5 (highest) according to commands "20K" .. "25K".
// Applied in reverse order if HEATER_PWM_REVERSE indicates reverse PWM signal.
// A Soft-PWM is applied, if HEATER_PWM is not defined.
// 
// Adjust appropriate to hardware response, e.g. P = V²/R:
//  Power 20..100% = 20, 40, 60, 80, 100% --> Voltage = sqrt(20%, 40%, 60%, 80%, 100%)% --> 
//
#if defined(HEATER_PWM) || defined(HEATER_SOFT_PWM)
const int HeaterPWMLevels[5] = {114, 161, 198, 228, 255};
#endif
  
/////////////////////////////////////////////////////////////////////////
//
//  Definitions and code needed to configure the temperature sensor 
//		and PID controller(s).
//
/////////////////////////////////////////////////////////////////////////
#if defined(EOS2000D) && defined(MY_2000D_2)   // 2000D w/ Bluetooth and cooling control

// Debug-Message, displayed after resetting the Arduino (can be omitted):
#define HELLO_MSG "2000D-2 mit Kühlung und Temperatursensor"

// Suggested PID constants (taking into account that ReadTemperatureX10() reads T times 10):
#define PID_Kp	-10	 // (increasing the cooling power by 10 PWM counts will lower(!) the sensor reading by 10 * 0.1°C )
#define PID_Ki	-0.1
#define PID_Kd	0        // or -10, but not very usefull (partially due to the low resolution of T measurement)
//
// NTC Thermistor Data
//
// The temperature is approximated after Steinhart–Hart with
// additional quadratic term from the NTC resistance R by:
// 
//        T = (1/(k0 + k1*x + k2*x² + k3*x³) - 273.15) °C
// 
// where x = ln(R / 10kOhm). Thereby, the division by 10kOhm is
//  chosen to match the effect of an 10k resistor in series (see below).
// The coefficients k0..k3 are determined from measurement to:
//
# define NTC_K0 0.0033743
# define NTC_K1 0.00027042
# define NTC_K2 -5.9567e-05
# define NTC_K3 9.4746e-05

// 
// Read the temperature sensor
// 
// In my 2000D-2, an NTC near the sensor is fed from Vcc with a
//  10k resistor in series. So Vcc is scaled by R / (R + 10 kOhm) = 1/(1 + 10k/R).
// With analogReference(DEFAULT) executed in setupTemperatureControl(),
//  1023 corresponds to Vin == Vcc.
// --> R = 10 kOhm / (1023/V_reading - 1)
// 
// Attention: Do not call during a shutter action, since a
//  reading will take 5 to 20 microseconds (and more if the clock
//  frequency is below 16 MHz) and I didn't check whether shutter
//  signals will still be in time.
// 
// Return 1 / T(K).
// 
#pragma GCC push_options
#pragma GCC optimize ("-O3")   // replace -Os (optimize size) --> smaller!
//inline float ReadTemperature_times_10() __attribute__((optimize("-O3")));

inline float Read_1_T()
{
  int V_reading;
  //*
  V_reading = analogRead(TEMPSENS);
  /*/
  // Testing w/o hardware:
  static int ti = 0;
  const float readings[4] = {0.5, 0.505, 0.6, 0.4}; // V @ 25°C, 24.5°C, .. of perfect B = 3950K NTC
  V_reading = int(readings[ti%4] * 1023);
  ti++;
  //*/
  // Calculate R / 10kOhm = 1 / (1023/V_reading - 1) = V_reading / (1023 - V_reading)
  // (
  // 
  if (V_reading == 0) {             // should not happen
    return 999;
  }
  float R_10k = V_reading / (float)(1023 - V_reading);
  float x = log(R_10k) - 0.01;
  // -0.01 is an empirical correction of the resistance by -1%,
  //  which increases T by +0.23°C.
  // This correction compensates for the tolerances of the NTC and 
  //  the deviding resistor, which are in the range of 1% typically.
  // An additional correction in the range of -0.1°C could be 
  //  applied to compensate for the self-heating of the NTC, which
  //  consumes about (3.3V / 2)^2 / 10kOhm ~= 0.27 mW. 
  //  (In air, 1 mW will increase T by 0.5 to 1°C.)

  return (((NTC_K3 * x + NTC_K2) * x + NTC_K1) * x + NTC_K0);	// 1/T

  // Simplified B parameter equation 
  //   1/T = 1/(25+273.15 K) + 1/B * ln(R/Ro)
  // 
  // with R(25°C) = 10kOhm and B = 3950K:
  //   1/T = 0.00335402 + 0.00025317 * ln(R/10kOhm)
  //
  //return 0.00335402 + x * 0.00025317;
  // (Does not fit very well with my particular NTC.)
}


//
// Return 10 * T(°C) to allow temperature being handled as an integer.
// 
// (At 25°C, a B=3950K NTC shows a relative change of resistance of 0.5% / °C,
//  while the resolution of a 10 Bit ADC is 0.1%. 
//  So a 10 Bit ADC is resolving 0.2°C / LSB.)
// 
inline float ReadTemperature_times_10()
{
  //return 10. / Read_1_T - 273.15*10 + 0.5;
  // looks like gcc doesn't optimize like this:
  return 10. / Read_1_T() - 2731;

  // Added 0.5 for rounding to int, but still returning as float,
  // because gcc will produce 60 extra bytes of code if we return
  // an int!!!
}
#pragma GCC pop_options

#endif



// Setup temperature control
//
inline void setupTemperatureControl()
{
#ifdef TEMPSENS
  analogReference(DEFAULT);     // use Vcc as Ref for ADC
#endif 
#ifdef COOLER_POWER
  pinMode(COOLER_POWER, OUTPUT);
  digitalWrite(COOLER_POWER, !COOLER_POWER_ON);
#endif 
#ifdef HEATER_POWER
  pinMode(HEATER_POWER, OUTPUT);
  digitalWrite(HEATER_POWER, !HEATER_POWER_ON);
#endif 
}

#endif
